import sys
from functools import partial

# Hack to work around PySide being imported from nowhere:
import qtpy
from loguru import logger
from qtpy.QtCore import Qt, QObject, QEvent
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QMenuBar, QShortcut, QMenu, QWidget, QAction, QActionGroup, QApplication

if "PySide.QtCore" in sys.modules and qtpy.API != "pyside":
    del sys.modules["PySide.QtCore"]

from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtconsole.inprocess import QtInProcessKernelManager


class DebuggableMenuBar(QMenuBar):
    def __init__(self, *args, **kwargs):
        super(DebuggableMenuBar, self).__init__(*args, **kwargs)

        self.debugshortcut = QShortcut(QKeySequence("Ctrl+Return"),
                                       self,
                                       self.showDebugMenu,
                                       context=Qt.ApplicationShortcut)

        self._debugmenu = QMenu("Debugging")
        self._debugmenu.addAction("Debug widget", self.startDebugging)
        self._loggingmenu = QMenu("Logging level")
        self._debugmenu.addMenu(self._loggingmenu)
        self._levels_group = QActionGroup(self)
        self._levels_group.setExclusive(True)
        for level in ['Critical', 'Error', 'Warning', 'Info', 'Debug']:
            action = QAction(level)
            action.triggered.connect(partial(self.set_level, level))
            action.setCheckable(True)
            self._levels_group.addAction(action)
            self._loggingmenu.addAction(action)
        self._loggingmenu.actions()[2].setChecked(True)

        self.mousedebugger = MouseDebugger()

    def set_level(self, level: str):
        from . import displays
        logger.remove(displays.log_handler_id)
        displays.log_handler_id = logger.add(displays.LogHandler(), level=level.upper())
        logger.critical(f'Log level set to {level.upper()}')

    def showDebugMenu(self):
        self.addMenu(self._debugmenu)

    def startDebugging(self):
        QApplication.instance().installEventFilter(self.mousedebugger)


class MouseDebugger(QObject):
    def eventFilter(self, obj, event):
        # print(event,obj)
        # print(self.sender())
        if event.type() == QEvent.MouseButtonPress:
            print(QApplication.instance().activeWindow().childAt(event.pos()))
            IPythonDebugger(QApplication.instance().activeWindow().childAt(event.pos())).show()
            QApplication.instance().removeEventFilter(self)
            return True
        return False


class IPythonDebugger(RichJupyterWidget):
    def __init__(self, widget: QWidget):
        super(IPythonDebugger, self).__init__()

        # Setup the kernel
        self.kernel_manager = QtInProcessKernelManager()
        self.kernel_manager.start_kernel()
        kernel = self.kernel_manager.kernel
        kernel.gui = "qt"

        # Push QWidget to the console
        kernel.shell.push({"widget": widget})

        self.kernel_client = self.kernel_manager.client()
        self.kernel_client.start_channels()

        # Setup console widget
        def stop():
            self.kernel_client.stop_channels()
            self.kernel_manager.shutdown_kernel()

        self.exit_requested.connect(stop)


if __name__ == "__main__":
    from qtpy.QtWidgets import QMainWindow, QLabel

    app = QApplication([])
    window = QMainWindow()
    window.setCentralWidget(QLabel("test"))
    db = DebuggableMenuBar()
    window.setMenuBar(db)
    window.show()

    app.exec_()



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import ClassVar
from uuid import uuid4

from qtpy.QtWidgets import QWidget
from qtpy.QtCore import Signal, QObject


class Location(Enum):
    Client = auto()
    Core = auto()
    ExecutionEngine = auto()
    AdaptiveEngine = auto()


class ComputeMode(Enum):
    Blocking = auto()
    Threaded = auto()


class RenderMode(Enum):
    Blocking = auto()
    Background = auto()


class GraphSignalRelay(QObject):
    sigPush = Signal(object)


graph_signal_relay = GraphSignalRelay()


@dataclass(eq=False)
class Graph:
    name: ClassVar[str] = ''
    compute_with: Location = Location.Client
    compute_mode: ComputeMode = ComputeMode.Blocking
    render_mode: RenderMode = RenderMode.Blocking
    widget_class: ClassVar[type[QWidget]] = None
    widget_args: tuple = field(default_factory=tuple)
    widget_kwargs: dict = field(default_factory=dict)
    id: str = field(default_factory=lambda: str(uuid4()))


    def compute(self, data: 'Data', *args):
        ...

    def update(self, data: 'Data', *args):
        ...

    def make_widget(self):
        return self.widget_class(*self.widget_args, **self.widget_kwargs)

    # @property
    # def widget(self):
    #     return
    #
    # @widget.setter
    # def widget(self, value):
    #     self._widget = value



import sys
import warnings
from collections.abc import Callable

import numpy as np

import pyqtgraph as pg
from pyqtgraph import debug, getConfigOption, functions as fn, colormap, ColorMap, getCupy, rescaleData, applyLookupTable
from pyqtgraph.Qt import QtCore, QtWidgets, QtGui
from pyqtgraph.graphicsItems.ScatterPlotItem import SymbolAtlas, ScatterPlotItem
from scipy.spatial import Delaunay

pg.setConfigOption('useOpenGL', True)
pg.setConfigOption('enableExperimental', True)


HAVE_OPENGL = hasattr(QtWidgets, 'QOpenGLWidget')


class CloudItem(pg.GraphicsObject):
    """
    An opengl point cloud visualization utilizing Delaunay Triangulation. Hardware acceleration is required.
    """

    sigDataChanged = sigImageChanged = QtCore.Signal()
    sigClicked = QtCore.Signal(object, object)
    sigPointsClicked = QtCore.Signal(object, object, object)
    sigPointsHovered = QtCore.Signal(object, object, object)

    def __init__(self, **kwargs):
        super(CloudItem, self).__init__(kwargs.get('parent', None))
        self.levels = None
        self.clear()

        self.scatter = ScatterPlotItem(**kwargs)
        self.scatter.setParentItem(self)
        self.scatter.sigClicked.connect(self.scatterClicked)
        self.scatter.sigHovered.connect(self.scatterHovered)

        self.metaData = {}
        self.opts = {
            'name': None,
            'antialias': getConfigOption('antialias'),
            'mouseWidth': 8, # width of shape responding to mouse click
            'compositionMode': None,
            'pen': None,
            'hoverable': False,
            'tip': None,
        }

        # self.setClickable(kwargs.get('clickable', False))

        if 'x' in kwargs:
            self.setData(**kwargs)

    def setData(self, **kwargs):
        """
        =============== =================================================================
        **Arguments:**
        x, y            (numpy arrays) Data positions to display
        c               (numpy array) Data values to display
        pen             Pen to use when drawing. Any single argument accepted by
                        :func:`mkPen <pyqtgraph.mkPen>` is allowed.
        antialias       (bool) Whether to use antialiasing when drawing. This
                        is disabled by default because it decreases performance.
        compositionMode See :func:`setCompositionMode
                        <pyqtgraph.PlotCurveItem.setCompositionMode>`.
        colorMap :      pg.ColorMap, default pg.colormap.get('viridis')
                        Colormap used to map the z value to colors.
        *hoverable*            If True, sigHovered is emitted with a list of hovered points, a tool tip is shown containing
                        information about them, and an optional separate style for them is used. Default is False.
        *tip*           A string-valued function of a spot's (x, y, data) values. Set to None to prevent a tool tip
                        from being shown.
        *hoverSymbol*   A single symbol to use for hovered spots. Set to None to keep symbol unchanged. Default is None.
        *hoverSize*     A single size to use for hovered spots. Set to -1 to keep size unchanged. Default is -1.
        *hoverPen*      A single pen to use for hovered spots. Set to None to keep pen unchanged. Default is None.
        *hoverBrush*    A single brush to use for hovered spots. Set to None to keep brush unchanged. Default is None.
        =============== =================================================================

        **Notes on performance:**

        Line widths greater than 1 pixel affect the performance as discussed in
        the documentation of :class:`PlotDataItem <pyqtgraph.PlotDataItem>`.
        """
        if 'autoLevels' not in kwargs:
            kwargs['autoLevels'] = True
        self.updateData(**kwargs)

    def updateData(self, **kwargs):
        self.clear()
        self.scatter.clear()
        self.extendData(**kwargs)

    def extendData(self, x, y, c, autoLevels=False, **kwargs):
        profiler = debug.Profiler()

        if 'compositionMode' in kwargs:
            self.setCompositionMode(kwargs['compositionMode'])
        if 'hoverable' in kwargs:
            self.opts['hoverable'] = bool(kwargs['hoverable'])
        if 'tip' in kwargs:
            self.opts['tip'] = kwargs['tip']

        data = {'x': x, 'y': y, 'c': c}

        for k, datum in data.items():
            if isinstance(datum, (list, tuple)):
                data[k] = datum = np.array(datum)
            if not isinstance(datum, np.ndarray) or datum.ndim > 1:
                raise Exception("Plot data must be 1D ndarray.")
            if datum.dtype.kind == 'c':
                raise Exception("Can not plot complex data types.")

        profiler("data checks")

        self.xData = np.append(self.xData, data['x'])
        self.yData = np.append(self.yData, data['y'])
        self.cData = np.append(self.cData, data['c'])

        if len(self.xData) > 4:
            if not self.delaunay:
                points = np.column_stack([self.xData, self.yData])
                self.delaunay = Delaunay(points, incremental=True)
            else:
                points = np.column_stack([data['x'], data['y']])
                self.delaunay.add_points(points)

            self.simplices = self.delaunay.simplices

        self.invalidateBounds()
        self.prepareGeometryChange()
        self.informViewBoundsChanged()

        profiler('copy')

        self._mouseShape = None

        if 'name' in kwargs:
            self.opts['name'] = kwargs['name']
        if 'pen' in kwargs:
            self.setPen(kwargs['pen'])
        if 'antialias' in kwargs:
            self.opts['antialias'] = kwargs['antialias']
        ## if symbol pen/brush are given with no previously set symbol, then assume symbol is 'o'
        if 'symbol' not in kwargs and ('symbolPen' in kwargs or 'symbolBrush' in kwargs or 'symbolSize' in kwargs):
            if self.opts['symbol'] is None:
                kwargs['symbol'] = 'o'
        if autoLevels is None:
            if 'levels' in kwargs:
                autoLevels = False
            else:
                autoLevels = True
        if autoLevels or self.levels is None:
            mn, mx = self.cData.min(), self.cData.max()
            # mn and mx can still be NaN if the data is all-NaN
            if mn == mx or np.isnan(mn) or np.isnan(mx):
                mn = 0
                mx = 255
            kwargs['levels'] = [mn, mx]
        if 'levels' in kwargs:
            self.setLevels(kwargs['levels'])

        self.scatter.addPoints(x=x, y=y, **kwargs)

        profiler('set')
        self.update()
        profiler('update')
        self.sigDataChanged.emit()
        profiler('emit')

    def implements(self, interface=None):
        ints = ['plotData']
        if interface is None:
            return ints
        return interface in ints

    def name(self):
        """ Returns the name that represents this item in the legend. """
        return self.opts.get('name', None)

    def setAlpha(self, alpha, auto):
        if self.opts['alphaHint'] == alpha and self.opts['alphaMode'] == auto:
            return
        self.opts['alphaHint'] = alpha
        self.opts['alphaMode'] = auto
        self.setOpacity(alpha)

    def setPen(self, *args, **kargs):
        """
        Sets the pen used to draw lines between points.
        The argument can be a :class:`QtGui.QPen` or any combination of arguments accepted by
        :func:`pyqtgraph.mkPen() <pyqtgraph.mkPen>`.
        """
        pen = fn.mkPen(*args, **kargs)
        self.opts['pen'] = pen
        #self.curve.setPen(pen)
        #for c in self.curves:
        #c.setPen(pen)
        #self.update()
        self.updateItems(styleUpdate=True)

    def setSymbol(self, symbol):
        """ `symbol` can be any string recognized by
        :class:`ScatterPlotItem <pyqtgraph.ScatterPlotItem>` or a list that
        specifies a symbol for each point.
        """
        if self.opts['symbol'] == symbol:
            return
        self.opts['symbol'] = symbol
        #self.scatter.setSymbol(symbol)
        self.updateItems(styleUpdate=True)

    def setSymbolPen(self, *args, **kargs):
        """
        Sets the :class:`QtGui.QPen` used to draw symbol outlines.
        See :func:`mkPen() <pyqtgraph.functions.mkPen>`) for arguments.
        """
        pen = fn.mkPen(*args, **kargs)
        if self.opts['symbolPen'] == pen:
            return
        self.opts['symbolPen'] = pen
        #self.scatter.setSymbolPen(pen)
        self.updateItems(styleUpdate=True)

    def setSymbolBrush(self, *args, **kargs):
        """
        Sets the :class:`QtGui.QBrush` used to fill symbols.
        See :func:`mkBrush() <pyqtgraph.functions.mkBrush>`) for arguments.
        """
        brush = fn.mkBrush(*args, **kargs)
        if self.opts['symbolBrush'] == brush:
            return
        self.opts['symbolBrush'] = brush
        #self.scatter.setSymbolBrush(brush)
        self.updateItems(styleUpdate=True)

    def setSymbolSize(self, size):
        """
        Sets the symbol size.
        """
        if self.opts['symbolSize'] == size:
            return
        self.opts['symbolSize'] = size
        #self.scatter.setSymbolSize(symbolSize)
        self.updateItems(styleUpdate=True)

    @debug.warnOnException  ## raising an exception here causes crash
    def paint(self, p, opt, widget):
        profiler = debug.Profiler()
        if self.xData is None or len(self.xData) == 0:
            return

        if getConfigOption('enableExperimental'):
            if HAVE_OPENGL and isinstance(widget, QtWidgets.QOpenGLWidget):
                self.paintGL(p, opt, widget)
                return
        raise RuntimeError('OpenGL and experimental mode must be enabled to use CloudItem')

    def paintGL(self, p, opt, widget):
        profiler = debug.Profiler()
        cData = self.cData
        levels = self.levels

        # Request a lookup table if this image has only one channel
        self.lut = self._ensure_proper_substrate(self.lut, np)
        if isinstance(self.lut, Callable):
            lut = self._ensure_proper_substrate(self.lut(self.cData), np)
        else:
            lut = self.lut

        if lut is None:
            lut = np.tile(range(256), [3, 1]).T

        if lut is not None:
            scale = lut.shape[0]
        else:
            scale = 255.

        if levels is not None:
            minVal, maxVal = levels
        else:
            minVal, maxVal = cData.min(), cData.max()

        if minVal == maxVal:
            maxVal = np.nextafter(maxVal, 2*maxVal)
        rng = maxVal-minVal
        rng = 1 if rng == 0 else rng

        cData = rescaleData(cData, scale / rng, minVal, dtype=cData.dtype)

        profiler('apply levels')

        cData = applyLookupTable(cData, lut).astype(np.ubyte)

        profiler('apply lut')


        p.beginNativePainting()
        import OpenGL.GL as gl

        if sys.platform == 'win32':
            # If Qt is built to dynamically load OpenGL, then the projection and
            # modelview matrices are not setup.
            # https://doc.qt.io/qt-6/windows-graphics.html
            # https://code.woboq.org/qt6/qtbase/src/opengl/qopenglpaintengine.cpp.html
            # Technically, we could enable it for all platforms, but for now, just
            # enable it where it is required, i.e. Windows
            gl.glMatrixMode(gl.GL_PROJECTION)
            gl.glLoadIdentity()
            gl.glOrtho(0, widget.width(), widget.height(), 0, -999999, 999999)
            gl.glMatrixMode(gl.GL_MODELVIEW)
            mat = QtGui.QMatrix4x4(self.sceneTransform())
            gl.glLoadMatrixf(np.array(mat.data(), dtype=np.float32))

        ## set clipping viewport
        view = self.getViewBox()
        if view is not None:
            rect = view.mapRectToItem(self, view.boundingRect())
            #gl.glViewport(int(rect.x()), int(rect.y()), int(rect.width()), int(rect.height()))

            #gl.glTranslate(-rect.x(), -rect.y(), 0)

            gl.glEnable(gl.GL_STENCIL_TEST)
            gl.glColorMask(gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE, gl.GL_FALSE) # disable drawing to frame buffer
            gl.glDepthMask(gl.GL_FALSE)  # disable drawing to depth buffer
            gl.glStencilFunc(gl.GL_NEVER, 1, 0xFF)
            gl.glStencilOp(gl.GL_REPLACE, gl.GL_KEEP, gl.GL_KEEP)

            ## draw stencil pattern
            gl.glStencilMask(0xFF)
            gl.glClear(gl.GL_STENCIL_BUFFER_BIT)
            gl.glBegin(gl.GL_TRIANGLES)
            gl.glVertex2f(rect.x(), rect.y())
            gl.glVertex2f(rect.x()+rect.width(), rect.y())
            gl.glVertex2f(rect.x(), rect.y()+rect.height())
            gl.glVertex2f(rect.x()+rect.width(), rect.y()+rect.height())
            gl.glVertex2f(rect.x()+rect.width(), rect.y())
            gl.glVertex2f(rect.x(), rect.y()+rect.height())
            gl.glEnd()

            gl.glColorMask(gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE)
            gl.glDepthMask(gl.GL_TRUE)
            gl.glStencilMask(0x00)
            gl.glStencilFunc(gl.GL_EQUAL, 1, 0xFF)

        try:
            # c = self.cData - min(self.cData)
            # c = np.nan_to_num(c / max(c) * (len(self._lut)-1))
            # lut_c = lut[cData.astype(np.int_)]
            gl.glBegin(gl.GL_TRIANGLES)
            for i, (i, j, k) in enumerate(self.simplices):
                ci, cj, ck = cData[[i, j, k]]
                gl.glColor3ub(*ci)
                gl.glVertex3f(self.xData[i], self.yData[i], 0)
                gl.glColor3ub(*cj)
                gl.glVertex3f(self.xData[j], self.yData[j], 0)
                gl.glColor3ub(*ck)
                gl.glVertex3f(self.xData[k], self.yData[k], 0)
            gl.glEnd()
        finally:
            p.endNativePainting()

    def clear(self):
        self.xData = np.array([])  ## raw values
        self.yData = np.array([])
        self.cData = np.array([])
        self.lut = None
        self._mouseShape = None
        self._mouseBounds = None
        self._boundsCache = [None, None]
        self._boundingRect = None
        self.fragmentAtlas = SymbolAtlas()
        self.delaunay = None
        self.simplices = []
        self._effectiveLut = None

    def boundingRect(self):
        if self._boundingRect is None:
            (xmn, xmx) = self.dataBounds(ax=0)
            if xmn is None or xmx is None:
                return QtCore.QRectF()
            (ymn, ymx) = self.dataBounds(ax=1)
            if ymn is None or ymx is None:
                return QtCore.QRectF()

            px = py = 0.0
            pxPad = self.pixelPadding()
            if pxPad > 0:
                # determine length of pixel in local x, y directions
                px, py = self.pixelVectors()
                try:
                    px = 0 if px is None else px.length()
                except OverflowError:
                    px = 0
                try:
                    py = 0 if py is None else py.length()
                except OverflowError:
                    py = 0

                # return bounds expanded by pixel size
                px *= pxPad
                py *= pxPad
            #px += self._maxSpotWidth * 0.5
            #py += self._maxSpotWidth * 0.5
            self._boundingRect = QtCore.QRectF(xmn-px, ymn-py, (2*px)+xmx-xmn, (2*py)+ymx-ymn)

        return self._boundingRect

    def dataBounds(self, ax, frac=1.0, orthoRange=None):
        ## Need this to run as fast as possible.
        ## check cache first:
        cache = self._boundsCache[ax]
        if cache is not None and cache[0] == (frac, orthoRange):
            return cache[1]

        if self.xData is None or len(self.xData) == 0:
            return (None, None)

        if ax == 0:
            d = self.xData
            d2 = self.yData
        elif ax == 1:
            d = self.yData
            d2 = self.xData
        else:
            raise ValueError("Invalid axis value")

        ## If an orthogonal range is specified, mask the data now
        if orthoRange is not None:
            mask = (d2 >= orthoRange[0]) * (d2 <= orthoRange[1])
            d = d[mask]
            #d2 = d2[mask]

        if len(d) == 0:
            return (None, None)

        ## Get min/max (or percentiles) of the requested data range
        if frac >= 1.0:
            # include complete data range
            # first try faster nanmin/max function, then cut out infs if needed.
            with warnings.catch_warnings():
                # All-NaN data is acceptable; Explicit numpy warning is not needed.
                warnings.simplefilter("ignore")
                b = (np.nanmin(d), np.nanmax(d))
            if np.isinf(b[0]) or np.isinf(b[1]):
                mask = np.isfinite(d)
                d = d[mask]
                if len(d) == 0:
                    return (None, None)
                b = (d.min(), d.max())

        elif frac <= 0.0:
            raise Exception("Value for parameter 'frac' must be > 0. (got %s)" % str(frac))
        else:
            # include a percentile of data range
            mask = np.isfinite(d)
            d = d[mask]
            if len(d) == 0:
                return (None, None)
            b = np.percentile(d, [50 * (1 - frac), 50 * (1 + frac)])

        ## Add pen width only if it is non-cosmetic.
        # pen = self.opts['pen']
        # spen = self.opts['shadowPen']
        # if not pen.isCosmetic():
        #     b = (b[0] - pen.widthF()*0.7072, b[1] + pen.widthF()*0.7072)
        # if spen is not None and not spen.isCosmetic() and spen.style() != QtCore.Qt.PenStyle.NoPen:
        #     b = (b[0] - spen.widthF()*0.7072, b[1] + spen.widthF()*0.7072)

        self._boundsCache[ax] = [(frac, orthoRange), b]
        return b

    def pixelPadding(self):
        # pen = self.opts['pen']
        # spen = self.opts['shadowPen']
        w = 0
        # if pen.isCosmetic():
        #     w += pen.widthF()*0.7072
        # if spen is not None and spen.isCosmetic() and spen.style() != QtCore.Qt.PenStyle.NoPen:
        #     w = max(w, spen.widthF()*0.7072)
        # if self.clickable:
        #     w = max(w, self.opts['mouseWidth']//2 + 1)
        return w

    def invalidateBounds(self):
        self._boundingRect = None
        self._boundsCache = [None, None]

    def scatterClicked(self, plt, points, ev):
        self.sigClicked.emit(self, ev)
        self.sigPointsClicked.emit(self, points, ev)

    def scatterHovered(self, plt, points, ev):
        self.sigPointsHovered.emit(self, points, ev)

    def setLookupTable(self, lut, update=True):
        """
        Sets lookup table ``lut`` to use for false color display of a monochrome image. See :func:`makeARGB <pyqtgraph.makeARGB>` for more
        information on how this is used. Optionally, `lut` can be a callable that accepts the current image as an
        argument and returns the lookup table to use.

        Ordinarily, this table is supplied by a :class:`~pyqtgraph.HistogramLUTItem`,
        :class:`~pyqtgraph.GradientEditorItem` or :class:`~pyqtgraph.ColorBarItem`.

        Setting ``update = False`` avoids an immediate image update.
        """
        if lut is not self.lut:
            lut = self._ensure_proper_substrate(lut, np)
            self.lut = lut
            self._effectiveLut = None
            if update:
                self.update()

    def setLevels(self, levels, update=True):
        """
        Sets image scaling levels.
        See :func:`makeARGB <pyqtgraph.makeARGB>` for more details on how levels are applied.

        Parameters
        ----------
            levels: list_like
                - ``[blackLevel, whiteLevel]``
                  sets black and white levels for monochrome data and can be used with a lookup table.
                - ``[[minR, maxR], [minG, maxG], [minB, maxB]]``
                  sets individual scaling for RGB values. Not compatible with lookup tables.
            update: bool, optional
                Controls if image immediately updates to reflect the new levels.
        """

        if levels is not None:
            levels = np.asarray(levels)
        self.levels = levels
        self._effectiveLut = None
        if update:
            self.update()

    @staticmethod
    def _ensure_proper_substrate(data, substrate):
        if data is None or isinstance(data, Callable) or isinstance(data, substrate.ndarray):
            return data
        cupy = getCupy()
        if substrate == cupy and not isinstance(data, cupy.ndarray):
            data = cupy.asarray(data)
        elif substrate == np:
            if cupy is not None and isinstance(data, cupy.ndarray):
                data = data.get()
            else:
                data = np.asarray(data)
        return data

    def getHistogram(self, bins='auto', step='auto', perChannel=False, targetImageSize=200,
                     targetHistogramSize=500, **kwds):
        """
        Returns `x` and `y` arrays containing the histogram values for the current image.
        For an explanation of the return format, see :func:`numpy.histogram()`.

        The `step` argument causes pixels to be skipped when computing the histogram to save time.
        If `step` is 'auto', then a step is chosen such that the analyzed data has
        dimensions approximating `targetImageSize` for each axis.

        The `bins` argument and any extra keyword arguments are passed to
        :func:`self.xp.histogram()`. If `bins` is `auto`, a bin number is automatically
        chosen based on the image characteristics:

          * Integer images will have approximately `targetHistogramSize` bins,
            with each bin having an integer width.
          * All other types will have `targetHistogramSize` bins.

        If `perChannel` is `True`, then a histogram is computed for each channel,
        and the output is a list of the results.
        """
        # This method is also used when automatically computing levels.
        if self.cData is None or len(self.cData) == 0:
            return None, None
        if step == 'auto':
            step = max(1, int(np.ceil(len(self.cData) / targetImageSize)))
        stepData = self.cData[::step]

        if isinstance(bins, str) and bins == 'auto':
            mn = np.nanmin(stepData).item()
            mx = np.nanmax(stepData).item()
            if mx == mn:
                # degenerate image, arange will fail
                mx += 1
            if np.isnan(mn) or np.isnan(mx):
                # the data are all-nan
                return None, None
            if stepData.dtype.kind in "ui":
                # For integer data, we select the bins carefully to avoid aliasing
                step = int(np.ceil((mx - mn) / targetHistogramSize))
                bins = []
                if step > 0.0:
                    bins = np.arange(mn, mx + 1.01 * step, step, dtype=int)
            else:
                # for float data, let numpy select the bins.
                bins = np.linspace(mn, mx, targetHistogramSize)

            if len(bins) == 0:
                bins = np.asarray((mn, mx))

        kwds['bins'] = bins

        cp = getCupy()
        if perChannel:
            hist = []
            for i in range(stepData.shape[-1]):
                stepChan = stepData[..., i]
                stepChan = stepChan[np.isfinite(stepChan)]
                h = np.histogram(stepChan, **kwds)
                if cp:
                    hist.append((cp.asnumpy(h[1][:-1]), cp.asnumpy(h[0])))
                else:
                    hist.append((h[1][:-1], h[0]))
            return hist
        else:
            stepData = stepData[np.isfinite(stepData)]
            hist = np.histogram(stepData, **kwds)
            if cp:
                return cp.asnumpy(hist[1][:-1]), cp.asnumpy(hist[0])
            else:
                return hist[1][:-1], hist[0]

    def getLevels(self):
        return self.levels


import enum


class _SimpleReprEnum(enum.Enum):
    def __repr__(self):
        return self.name


class Role(_SimpleReprEnum):
    CLIENT = enum.auto()
    SERVER = enum.auto()


class Direction(_SimpleReprEnum):
    RESPONSE = enum.auto()
    REQUEST = enum.auto()


Commands = {}
Commands[Role.CLIENT] = {}
Commands[Role.SERVER] = {}
_commands = set()

ENCODING = 'ascii'


class _MetaDirectionalMessage(type):
    # see what you've done, @tacaswell and @danielballan?
    def __new__(metacls, name, bases, dct):
        if name.endswith('Request'):
            direction = Direction.REQUEST
            command_dict = Commands[Role.CLIENT]
        else:
            direction = Direction.RESPONSE
            command_dict = Commands[Role.SERVER]

        dct['DIRECTION'] = direction
        new_class = super().__new__(metacls, name, bases, dct)

        if new_class.FNC is not None:
            command_dict[new_class.FNC] = new_class

        if not name.startswith('_'):
            _commands.add(new_class)
        return new_class


class Message(metaclass=_MetaDirectionalMessage):
    __slots__ = ('payload',)
    WRITE_REQUIRED = False
    FNC = None

    def __init__(self, *payload):
        self.payload = payload
        for name, value in zip(self.__slots__, payload):
            setattr(self, name, value)

    # @classmethod
    # def from_wire(cls, payload_buffers):
    #     return cls.from_components(
    #         str(b''.join(payload_buffers), ENCODING).strip())

    # @classmethod
    # def from_components(cls, str_payload):
    #     # Bwahahahaha
    #     instance = cls.__new__(cls)
    #     instance.str_payload = str_payload
    #     return instance

    # def __bytes__(self):
    #     return bytes(self.str_payload, ENCODING)

    def __repr__(self):
        return f'{self.__class__.__name__}: {self.payload!r}'


class _DataResponse(Message):
    def __repr__(self):
        return f'{self.__class__.__name__} size: {len(self.payload[0]["positions"])}'


class PushDataRequest(Message):
    __slots__ = ('data',)


class PushDataResponse(Message):
    __slots__ = ()


class FullDataRequest(Message):
    __slots__ = ()


class FullDataResponse(_DataResponse):
    __slots__ = ('data',)


class PartialDataRequest(Message):
    __slots__ = ('iteration',)


class PartialDataResponse(_DataResponse):
    __slots__ = ('data', 'last_data_size')


class StartRequest(Message):
    __slots__ = ()


class StopRequest(Message):
    __slots__ = ()


class PauseRequest(Message):
    __slots__ = ()


class ExitRequest(Message):
    __slots__ = ()


class StateRequest(Message):
    __slots__ = ()


class StateResponse(Message):
    __slots__ = ('state', 'compute_metrics')


class GetParametersRequest(Message):
    __slots__ = ()


class GetParametersResponse(Message):
    __slots__ = ('parameter_state',)


class SetParameterRequest(Message):
    __slots__ = ('child_path', 'value')


class SetParameterResponse(Message):
    __slots__ = ('success',)


class UnknownResponse(Message):
    __slots__ = ()


class ExceptionResponse(Message):
    __slots__ = ('exception',)


class MeasureRequest(Message):
    __slots__ = ('position',)


class MeasureResponse(Message):
    __slots__ = ('success',)


class ConnectRequest(Message):
    __slots__ = ()


class ConnectResponse(Message):
    __slots__ = ('state', 'compute_metrics')


class PushGraphsRequest(Message):
    __slots__ = ('graphs',)


class PullGraphsRequest(Message):
    __slots__ = ()


class GraphsResponse(Message):
    __slots__ = ('graphs',)


class ReplayRequest(Message):
    __slots__ = ('positions', 'measurements')


class ReplayResponse(Message):
    __slots__ = ('enable',)


class SetComputeMetricsRequest(Message):
    __slots__ = ('compute_metrics',)


import os

import numpy as np
import pyqtgraph as pg
import pytest
import scipy.misc

from tsuchinoko.graphics_items.clouditem import CloudItem

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="Test doesn't work in Github Actions.")
def test_cloud(qtbot, monkeypatch):
    pg.setConfigOption('useOpenGL', True)
    pg.setConfigOption('enableExperimental', True)

    # Create window with GraphicsView widget
    win = pg.GraphicsLayoutWidget()
    win.show()  # show widget alone in its own window
    win.setWindowTitle('CloudItem Example')
    view = win.addViewBox()

    # image = np.asarray(Image.open('test.jpeg'))
    image = scipy.misc.ascent()
    x, y = np.random.random((2, 10000))
    x *= image.shape[1]
    y *= image.shape[0]
    c = [np.average(image[-int(yi), int(xi)]) for xi, yi in zip(x, y)]
    x, y = list(x), list(y)

    cloud = CloudItem(size=1)
    histlut = pg.HistogramLUTWidget()
    histlut.setImageItem(cloud)

    n = 1000
    cloud.extendData([x.pop() for i in range(n)], [y.pop() for i in range(n)], [c.pop() for i in range(n)])

    view.addItem(cloud)

    qtbot.wait_exposed(win)
    qtbot.wait_exposed(histlut)

    with monkeypatch.context() as m:
        with qtbot.waitCallback() as cb:
            # monkeypatch paint method to capture call
            def paintGL(*args, **kwargs):
                ret = CloudItem.paintGL(cloud, *args, **kwargs)
                cb()
                return ret

            m.setattr(cloud, 'paintGL', paintGL)

    qtbot.add_widget(win)
    qtbot.add_widget(histlut)



from pathlib import Path
from threading import Thread

import pytest
import numpy as np
from PIL import Image
from bluesky.plan_stubs import checkpoint, mov, trigger_and_read
from loguru import logger
from ophyd import Device
from ophyd.sim import SynAxis, SynSignal, Cpt
from pytest import fixture
from pytest_lazyfixture import lazy_fixture
from scipy import ndimage

from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.adaptive.random_in_process import RandomInProcess
from tsuchinoko.core import CoreState, ZMQCore
from tsuchinoko.execution.bluesky_in_process import BlueskyInProcessEngine
from tsuchinoko.execution.simple import SimpleEngine
from tsuchinoko.execution.threaded_in_process import ThreadedInProcessEngine
from tsuchinoko.utils.runengine import get_run_engine
# Disable logging to console when running tests
# NOTE: it seems there is a bug between loguru and pytest; pytest tries to log to a tempfile, but closes it when finished
# NOTE: if loguru has a backlog of messages
# logger.remove()




@fixture
def image_data():
    # Load data from a jpg image to be used as a luminosity map
    image = np.flipud(
        np.asarray(Image.open(Path(__file__).parent.parent / 'tsuchinoko' / 'examples' / 'sombrero_pug.jpg')))
    luminosity = np.average(image, axis=2)
    return luminosity


@fixture
def image_func(image_data):
    # Bilinear sampling will be used to effectively smooth pixel edges in source data
    def bilinear_sample(pos):
        return pos, ndimage.map_coordinates(image_data, [[pos[1]], [pos[0]]], order=1)[0], 1, {}

    return bilinear_sample


@fixture
def simple_execution_engine(image_func):
    execution = SimpleEngine(measure_func=image_func)
    return execution


@fixture
def gpcam_engine(image_data):
    # Define a gpCAM adaptive engine with initial parameters
    adaptive = GPCAMInProcessEngine(dimensionality=2,
                                    parameter_bounds=[(0, image_data.shape[1]),
                                                      (0, image_data.shape[0])],
                                    hyperparameters=[255, 100, 100],
                                    hyperparameter_bounds=[(0, 1e5),
                                                           (0, 1e5),
                                                           (0, 1e5)])
    return adaptive


@fixture
def random_engine(image_data):
    adaptive = RandomInProcess(dimensionality=2,
                               parameter_bounds=[(0, image_data.shape[1]),
                                                 (0, image_data.shape[0])],
                               max_targets=100
                               )  # this engine is FAST, so best to limit it
    return adaptive


@fixture
def bluesky_execution_engine(image_func):
    class PointDetector(Device):
        motor1 = Cpt(SynAxis, name='motor1')
        motor2 = Cpt(SynAxis, name='motor2')
        value = Cpt(SynSignal, name='value')

        def __init__(self, name):
            super(PointDetector, self).__init__(name=name)
            self.value.sim_set_func(self.get_value)

        def get_value(self):
            return image_func([int(self.motor2.position), int(self.motor1.position)])

        def trigger(self, *args, **kwargs):
            return self.value.trigger(*args, **kwargs)

    # Instantiate a simulated device
    point_detector = PointDetector('point_detector')

    # Define a Bluesky Plan component for performing measurements at targets.
    # Note that this returns the measured value and variance
    def measure_target(target):
        yield from checkpoint()
        yield from mov(point_detector.motor1, target[0], point_detector.motor2, target[1])
        ret = (yield from trigger_and_read([point_detector]))
        return ret[point_detector.value.name]['value'], 2  # variance of 1

    # Define a Bluesky Plan component to get the current position of the device
    # (which may not necessarily be precisely the requested position)
    def get_position():
        yield from checkpoint()
        return point_detector.motor1.position, point_detector.motor2.position

    # Define an execution engine with the measurement and get_position functions
    execution = BlueskyInProcessEngine(measure_target, get_position)

    yield execution

    logger.info('starting bluesky engine teardown')
    RE = get_run_engine()
    RE.RE.halt()
    RE.process_queue_thread.requestInterruption()
    RE.process_queue_thread.wait()
    logger.info('bluesky engine teardown finished')


@fixture
def threaded_execution_engine(image_func):
    execution = ThreadedInProcessEngine(image_func)
    yield execution
    execution.exiting = True
    execution.measure_thread.join()


@fixture(params=[lazy_fixture('random_engine'),
                 lazy_fixture('gpcam_engine')])
def adaptive_test_engines(simple_execution_engine, request):
    adaptive_engine = request.param

    return adaptive_engine, simple_execution_engine


@fixture(params=[lazy_fixture('threaded_execution_engine'),
                 lazy_fixture('bluesky_execution_engine')])
def execution_test_engines(random_engine, request):
    execution_engine = request.param

    return random_engine, execution_engine


@fixture(params=[lazy_fixture('execution_test_engines'),
                 lazy_fixture('adaptive_test_engines')])
def core(request):
    adaptive_engine, execution_engine = request.param
    logger.info('starting setup')
    core = ZMQCore()
    core.set_adaptive_engine(adaptive_engine)
    core.set_execution_engine(execution_engine)
    server_thread = Thread(target=core.main)
    server_thread.start()
    core.state = CoreState.Starting
    logger.info('setup complete')

    yield core

    core.exit()
    server_thread.join()
    logger.info('teardown complete')



from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.core import ZMQCore
from tsuchinoko.execution.bluesky_adaptive import BlueskyAdaptiveEngine

# NOTE: REQUIRES UNRELEASED DATABROKER VERSIONS
# NOTE: To run this demo, you must also start a TsuchinokoAgent with Bluesky-Adaptive.
#       See tsuchinoko.execution.bluesky_adaptive for a primitive mocked demo.

bounds = [(0, 100)] * 2

# Define a gpCAM adaptive engine with initial parameters
adaptive = GPCAMInProcessEngine(dimensionality=2,
                                parameter_bounds=bounds,
                                hyperparameters=[255, 100, 100],
                                hyperparameter_bounds=[(0, 1e5),
                                                       (0, 1e5),
                                                       (0, 1e5)])

execution = BlueskyAdaptiveEngine()

# Construct a core server
core = ZMQCore()
core.set_adaptive_engine(adaptive)
core.set_execution_engine(execution)

if __name__ == '__main__':
    # Start the core server
    core.main()



from dataclasses import dataclass
from functools import partial, lru_cache, cached_property
import sys
from loguru import logger
from pyqtgraph.parametertree.parameterTypes import SimpleParameter
from yaml import load, Loader

import cv2

import fvgp
from fvgp.gp import GP

import numpy as np
from scipy import sparse, ndimage, fftpack, linalg
from gpcam.gp_optimizer import GPOptimizer
from tsuchinoko.adaptive import Data
from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.adaptive.grid import Grid
from tsuchinoko.core import ZMQCore, CoreState
from tsuchinoko.execution.replay import InterpolatingEngine
from tsuchinoko.execution.simple import SimpleEngine
from tsuchinoko.graphs import Location
from tsuchinoko.graphs.common import Table, Image, ImageViewBlendROI, image_grid, GPCamPosteriorMean, \
    GPCamHyperparameterPlot, Score, Variance, GPCamAcquisitionFunction
from tsuchinoko.utils import threads
from tsuchinoko.graphs.specialized import ReconstructionGraph, ProjectionOperatorGraph, \
    SinogramSpaceGPCamAcquisitionFunction, ProjectionMask, ReconHistogram, sirt
import PIL

from tsuchinoko.utils.zmq_queue import Queue_decision
from uuid import uuid4


session_id = uuid4()


def trapez(y,y0,w):
    return np.clip(np.minimum(y+1+w/2-y0, -y+1+w/2+y0),0,1)


def weighted_line(r0, c0, r1, c1, w, rmin=0, rmax=np.inf):
    # The algorithm below works fine if c1 >= c0 and c1-c0 >= abs(r1-r0).
    # If either of these cases are violated, do some switches.
    if abs(c1-c0) < abs(r1-r0):
        # Switch x and y, and switch again when returning.
        xx, yy, val = weighted_line(c0, r0, c1, r1, w, rmin=rmin, rmax=rmax)
        return (yy, xx, val)

    # At this point we know that the distance in columns (x) is greater
    # than that in rows (y). Possibly one more switch if c0 > c1.
    if c0 > c1:
        return weighted_line(r1, c1, r0, c0, w, rmin=rmin, rmax=rmax)

    # The following is now always < 1 in abs
    slope = (r1-r0) / (c1-c0)

    # Adjust weight by the slope
    w *= np.sqrt(1+np.abs(slope)) / 2

    # We write y as a function of x, because the slope is always <= 1
    # (in absolute value)
    x = np.arange(c0, c1+1, dtype=float)
    y = x * slope + (c1*r0-c0*r1) / (c1-c0)

    # Now instead of 2 values for y, we have 2*np.ceil(w/2).
    # All values are 1 except the upmost and bottommost.
    thickness = np.ceil(w/2)
    yy = (np.floor(y).reshape(-1,1) + np.arange(-thickness-1,thickness+2).reshape(1,-1))
    xx = np.repeat(x, yy.shape[1])
    vals = trapez(yy, y.reshape(-1,1), w).flatten()

    yy = yy.flatten()

    # Exclude useless parts and those outside of the interval
    # to avoid parts outside of the picture
    mask = np.logical_and.reduce((yy >= rmin, yy < rmax, vals > 0))

    return (yy[mask].astype(int), xx[mask].astype(int), vals[mask])

# def weighted_line(y0, x0, y1, x1):
#     line = np.zeros((l_det, l_det), dtype=np.uint8)
#     cv2.line(line, (int(x0), int(y0)), (int(x1), int(y1)), (255,), lineType=cv2.LINE_AA)
#     return line/255.


def projection_operator(x, phi, map_size, center=None, width=1, length=None):
    if not length:
        length = map_size * 2

    if not center:
        center = (map_size/2, map_size/2)

    # get a sampling over a slice through x, phi
    v = np.array([-np.cos(np.deg2rad(phi)), np.sin(np.deg2rad(phi))]) * (-x + map_size/2)
    c = np.array(center)
    f = np.array([-np.sin(np.deg2rad(phi)), -np.cos(np.deg2rad(phi))]) * length / 2

    f0 = v + c + f
    f1 = v + c - f

    x0, y0 = f0
    x1, y1 = f1

    # draw a 1px line (with AA) as a mask from f0 to f1
    # NOTE: this is correct when the beam width is the same as the x/y resolution; for other cases,
    # you could use scipy.ndimage.zoom()
    rows, cols, vals = weighted_line(y0, x0, y1, x1, w=0.25)
    xy_mask = (cols >= 0) & (cols < map_size) & (rows >= 0) & (rows < map_size)
    masked_cols, masked_rows, masked_vals = cols[xy_mask], rows[xy_mask], vals[xy_mask]
    (map_mask := np.zeros((map_size, map_size)))[masked_rows, masked_cols] = masked_vals
    # map_mask = weighted_line(y0,x0,y1,x1)

    return map_mask


def get_distance_matrix(x1, x2):
    """
    Function to calculate the pairwise distance matrix of
    points in x1 and x2.

    Parameters
    ----------
    x1 : np.ndarray
        Numpy array of shape (U x D).
    x2 : np.ndarray
        Numpy array of shape (V x D).

    Return
    ------
    distance matrix : np.ndarray
    """
    d = np.zeros((len(x1), len(x2)))
    for i in range(x1.shape[1]): d += (x1[:, i].reshape(-1, 1) - x2[:, i]) ** 2
    return np.sqrt(d)


def kernel(x1,x2,hps,obj):
    # print('kernel:', hps, x1, x2)
    # with log_time('kernel', cumulative_key='kernel'):
    A = obj.A
    #this kernel takes elements from the real space the prior covariance in sinogram space
    # newstyle
    # d = obj.get_distance_matrix(x1, x2)
    # oldstyle
    d = get_distance_matrix(x1, x2)
    k = hps[0] * np.exp(-d**2/hps[1])
    if len(x1) == len(x2) == len(obj.x_data):  # training
        kernel = linalg.block_diag(*[A @ k @ A.T]*n_sinograms)
        # kernel = A @ k @ A.T
    elif len(x1) == len(x2) != len(obj.x_data):  # prediction
        kernel = k
    elif len(x1) != len(x2):  # prediction
        kernel = np.repeat(A @ k, n_sinograms, 0)
        # kernel = A @ k
    return kernel


def noise(x,hps,obj):
    A = obj.A
    #this function takes elements from the real space and returns noise (co)variances in sinogram space
    n = A @ (np.zeros((len(x))) + 0.01)
    return np.diag(n)

# newstyle
# def mean_func(x,hps,obj):
# oldstyle
def mean_func(obj, x, hps):
    A = obj.A
    #this function takes elements from the real space and returns prior mean values in sinogram space
    # m = A @ np.zeros(len(x),)
    if len(x) == len(obj.x_data):
        m = np.repeat(A @ np.zeros(len(x)), n_sinograms)
    else:
        m = np.zeros((len(x),))
    return m


def cost(origin, x, arguments=None):
    print('cost:', origin, x)
    cost_x = 1 / 8  # cost is 1/velocity
    cost_phi = 1 / 30
    exposure_time = 1
    estimated_total_experiment_time = 60*60*1  # in seconds
    x_origin, phi_origin = origin
    cost = [(exposure_time + max(abs(ix - x_origin) * cost_x, abs(iphi - phi_origin) * cost_phi))/estimated_total_experiment_time for
            ix, iphi in x]
    return cost


last_variance = (-1, None)
last_mean = (-1, None)
last_recon = (-1, None)


def memoized_posterior_variance(x, gp):
    global last_variance
    if len(gp.y_data) != last_variance[0]:
        last_variance = len(gp.y_data), gp.posterior_covariance(x, variance_only=True)["v(x)"]
    return last_variance[1]


def memoized_posterior_mean(x, gp):
    global last_mean
    if len(gp.y_data) != last_mean[0]:
        last_mean = len(gp.y_data), gp.posterior_mean(x)["f(x)"]
    return last_mean[1]


def memoized_recon(gp):
    global last_recon
    if len(gp.y_data) != last_recon[0]:
        last_recon = len(gp.y_data), sirt(gp.y_data, gp.A, num_iterations=30)
        gp.last_recon = last_recon[1]
    return last_recon[1]


def variance(x, gp):
    x = x.reshape(-1, gp.input_dim)

    projections = [projection_operator(*x_i, l_det) for x_i in x]
    grid_positions = np.fliplr(image_grid(((0, l_det), (0, l_det)), (l_det, l_det)))  # use positions from gp here!!!!!!!

    variances = memoized_posterior_variance(grid_positions[:-1], gp)
    res_s = [projection.ravel()[:-1] * variances for projection in projections]
    res = np.sum(res_s, axis=1)
    return res

# TODO: look at res_s image; check why diagonals are preferred


# def variance(x, gp):
#     x = x.reshape(-1, gp.input_dim)
#
#     projections = [projection_operator(*x_i, l_det) for x_i in x]
#     x_reals = [np.asarray(np.nonzero(projection)).T for projection in projections]  # Each x_real is a collection of real-space points translated from the input point in sinogram space
#     x_coeffs = [projection[np.nonzero(projection)] for projection, x_real in zip(projections, x_reals)]
#
#     res_s = np.asarray([gp.posterior_covariance(x_real, variance_only=True)["v(x)"] for x_real, x_coeff in zip(x_reals, x_coeffs)])
#     res = np.sum(res_s, axis=1)
#     return res


def ucb(x, gp):
    x = x.reshape(-1, gp.input_dim)

    projections = [projection_operator(*x_i, l_det) for x_i in x]
    x_reals = [np.asarray(np.nonzero(projection)).T for projection in
               projections]  # Each x_real is a collection of real-space points translated from the input point in sinogram space
    x_coeffs = [projection[np.nonzero(projection)] for projection, x_real in zip(projections, x_reals)]

    v_s = np.asarray([gp.posterior_covariance(x_real, variance_only=True)["v(x)"] / x_coeff for x_real, x_coeff in
                        zip(x_reals, x_coeffs)])
    m_s = np.asarray([gp.posterior_mean(x_real)["f(x)"] / x_coeff for x_real, x_coeff in
                        zip(x_reals, x_coeffs)])

    v = np.average(v_s, axis=1)
    m = np.average(m_s, axis=1)

    return m + 3.0 * np.sqrt(v)


def binary_resolve(x, gp):
    x = x.reshape(-1, gp.input_dim)

    projections = [projection_operator(*x_i, l_det) for x_i in x]
    grid_positions = np.fliplr(image_grid(((0, l_det), (0, l_det)), (l_det, l_det)))

    v_s = memoized_posterior_variance(grid_positions[:-1], gp)
    m_s = memoized_posterior_mean(grid_positions[:-1], gp)
    recon = memoized_recon(gp)

    v = np.array([projection.ravel()[:-1] * v_s for projection in projections])

    # normalize m_s by projection length

    # d = np.average(np.abs(np.abs(m)-1), axis=1) + 3.0 * np.sqrt(v)
    low = ThresholdResolve.binary_low.value()
    high = ThresholdResolve.binary_high.value()
    mid = (high-low)/2 + low
    width = high-mid

    binary_levels_distance = np.abs(np.abs(m_s - mid) - width)
    m = np.array([projection.ravel()[:-1] * binary_levels_distance for projection in projections])
    cumulative_measurements = np.array([projection.ravel() * np.sum(gp.A, axis=0) for projection in projections])

    average_binary_distance = np.average(m, axis=1)
    average_cumulative_measurements = np.maximum(np.average(cumulative_measurements, axis=1), .0000001)


    d = average_binary_distance / average_cumulative_measurements

    return d


def bilinear_sample(pos, data):
    # pos = np.random.random((2,)) * np.array([32, 180])
    # print(f'measuring: x={pos[1]:.1f} y={pos[0]:.1f}')
    # return pos, [ndimage.map_coordinates(sino, [[pos[1]], [pos[0]]], order=1)[0] for sino in data.values()], [.001] * n_sinograms, {}
    A = projection_operator(*pos, l_x).reshape(l_x, l_x)
    return pos, [np.sum(map * A) for map in data.values()], [.00001] * n_sinograms, {}


def push_to_queue(pos):
    logger.critical(f'publishing: {pos}')
    decision_queue.publish([{'session': session_id, 'position': pos}])
    while True:
        measurement = decision_queue.get()
        if measurement[0]['session'] != session_id:
            logger.critical('Stale data received. Clearing Queue...')
        else:
            break
    return measurement[0]['value']


class BackgroundTraining(GPCAMInProcessEngine):
    def __init__(self, start_training_at:int = 40, *args, **kwargs):
        self.training_thread = None
        self.start_training_at = start_training_at
        super().__init__(*args, **kwargs)

    def train(self):
        if not self.training_thread or self.training_thread.done:
            if len(self.optimizer.y_data) >= self.start_training_at:
                # pull values from optimizer
                # newstyle
                # x, y, v, A = self.optimizer.x_data.copy(), self.optimizer.y_data.copy(), self.optimizer.V.copy(), self.optimizer.A.copy()
                # oldstyle
                x, y, v, A = self.optimizer.x_data.copy(), self.optimizer.y_data.copy(), self.optimizer.variances.copy(), self.optimizer.A.copy()

                # pull parameters
                hyperparameters_bounds = np.asarray([[self.parameters[('hyperparameters', f'hyperparameter_{i}_{edge}')]
                             for edge in ['min', 'max']]
                            for i in range(self.num_hyperparameters)])
                hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                            for i in range(self.num_hyperparameters)])
                parameter_bounds = np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                                                for edge in ['min', 'max']]
                                               for i in range(self.dimensionality)])

                self.training_thread = threads.QThreadFuture(self._background_train,
                                                             x,
                                                             y,
                                                             v,
                                                             A,
                                                             parameter_bounds,
                                                             hyperparameters,
                                                             hyperparameters_bounds,
                                                             {'method': 'global'})
                self.training_thread.start()

        return True

    def _background_train(self, x, y, v, A, parameter_bounds, hyperparameters, hyperparameter_bounds, training_kwargs):
        logger.info('Training asynchronously...')
        # newstyle
        # optimizer = GPOptimizer(x,
        #                         y,
        #                         noise_variances=v,
        #                         init_hyperparameters=hyperparameters,
        #                         hyperparameter_bounds=hyperparameter_bounds,
        #                         **self.gp_opts.copy())

        # oldstyle
        optimizer = GPOptimizer(self.dimensionality, parameter_bounds)
        optimizer.A = A.copy()
        optimizer.tell(x, y, v)
        opts = self.gp_opts.copy()
        optimizer.init_gp(init_hyperparameters=hyperparameters, **opts)
        optimizer.train(hyperparameter_bounds=hyperparameter_bounds, init_hyperparameters=hyperparameters, **training_kwargs)

        self.optimizer.hyperparameters = optimizer.hyperparameters
        logger.info(f'Hyperparameters set from asynchronous training: {optimizer.hyperparameters}')


class ProjectionOperatorBuilderGPCamEngine(BackgroundTraining):
    default_retrain_locally_at = tuple()
    default_retrain_globally_at = tuple() #82

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.graphs = [
            # GPCamPosteriorCovariance(),
            #            SinogramSpaceGPCamAcquisitionFunction(shape=(32, 32), real_space_bounds=((0, l_x), (0, l_x))),
                       ReconstructionGraph(shape=(l_x, l_x)),
                       ReconHistogram(),
                       ProjectionMask(shape=(l_x, l_x)),
                       ProjectionOperatorGraph(),
                       # GPCamAverageCovariance(),
                       GPCamHyperparameterPlot(),
                       Table(),
                       # Variance(),
                       Score()
        ]

    def init_optimizer(self):
        opts = self.gp_opts.copy()
        # TODO: only fallback to numpy when packaged as an app
        if sys.platform == 'darwin':
            opts['compute_device'] = 'numpy'




        # oldstyle
        parameter_bounds = np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                                        for edge in ['min', 'max']]
                                       for i in range(self.dimensionality)])
        self.optimizer = GPOptimizer(self.dimensionality, parameter_bounds)

        # newstyle
        # real_space_positions = np.mgrid[:l_x, :l_x].T.reshape(-1, 2) # swap in full real-space grid
        # self.optimizer = GPOptimizer(real_space_positions,
        #                              np.random.rand(2),
        #                              #noise_variances=np.ones((2,))*.01,
        #                              **opts)  # give hyperparameters!!!!!!!!!!!!!!!!!

    def update_measurements(self, data: Data):
        with data.r_lock():  # quickly grab values within lock before passing to optimizer
            positions = data.positions.copy()
            scores = data.scores.copy()
            variances = data.variances.copy()

        real_space_positions = np.mgrid[:l_x,:l_x].T.reshape(-1, 2)
        scores = np.array(scores).ravel()
        variances = np.array(variances).ravel()

        length_diff = len(positions) - self._positions_seen
        if length_diff:
            for position in positions[-length_diff:]:
                A_vec = projection_operator(*position, l_x).reshape(1, l_x **2)
                self.optimizer.A = np.vstack([self.optimizer.A, A_vec])
                self._positions_seen += 1
                # self.optimizer.positions = np.vstack([self.optimizer.positions, position])

        self.optimizer.tell(np.asarray(real_space_positions), np.asarray(scores), np.asarray(variances))

        # oldstyle
        if not self.optimizer.gp_initialized:
            hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                          for i in range(self.num_hyperparameters)])
            opts = self.gp_opts.copy()
            # TODO: only fallback to numpy when packaged as an app
            if sys.platform == 'darwin':
                opts['compute_device'] = 'numpy'

            self.init_gp(hyperparameters, **opts)

    def reset(self):
        super().reset()
        self.optimizer.A = np.zeros((0, l_x ** 2))
        self.optimizer.positions = np.zeros((0, 2))
        self._positions_seen = 0
        if self.graphs:
            self.graphs[0].last_recon = None


class EmptyOptimizer():
    pass


class ProjectionOperatorGrid(Grid):
    def __init__(self, *args, **kwargs):
        self.optimizer = EmptyOptimizer()
        super().__init__(*args, **kwargs)
        self.graphs = [ReconstructionGraph(),
                       Table(),
                       Score()
                       ]

    def update_measurements(self, data: Data):
        with data.r_lock():  # quickly grab values within lock before passing to optimizer
            positions = data.positions.copy()

        length_diff = len(positions) - self._positions_seen
        if length_diff:
            for position in positions[-length_diff:]:
                A_vec = projection_operator(*position, l_x).reshape(1, l_x **2)
                self.optimizer.A = np.vstack([self.optimizer.A, A_vec])
                self._positions_seen += 1

    def update_metrics(self, data: Data):
        for graph in self.graphs:
            try:
                graph.compute(data, self)
            except Exception as ex:
                logger.exception(ex)

    def reset(self):
        super().reset()
        self.optimizer.A = np.zeros((0, l_x ** 2))
        self._positions_seen = 0
        if self.graphs:
            self.graphs[0].last_recon = None


class GridFirst(ProjectionOperatorBuilderGPCamEngine):
    def __init__(self, grid_until_N:int, parameter_bounds, *args, **kwargs):
        self.grid_until_N = grid_until_N
        self.grid_engine = Grid(parameter_bounds=parameter_bounds)
        super().__init__(*args, parameter_bounds=parameter_bounds, **kwargs)

    def request_targets(self, position):
        if not hasattr(self.optimizer, 'y_data') or len(self.optimizer.y_data) < self.grid_until_N:
            return self.grid_engine.request_targets(position)
        else:
            return super().request_targets(position)

    def reset(self):
        super().reset()

        self.grid_engine.reset()


class SessionReset(GPCAMInProcessEngine):
    def reset(self):
        global session_id
        session_id = uuid4()
        super().reset()


class ThresholdResolve(GridFirst, SessionReset):
    binary_low = SimpleParameter(title='Binary Low', name='binary_low', type='float', value=-1)
    binary_high = SimpleParameter(title='Binary High', name='binary_high', type='float', value=1)
    @cached_property
    def parameters(self):
        parameters = super().parameters
        parameters.insertChild(0, self.binary_low)
        parameters.insertChild(0, self.binary_high)
        return parameters

n_angles = 180  # number of angles
l_det = l_x = 32  # sampling for ground truth sinograms

# #### FOR KANUPRIYA's ####
# geom2d = Geometry2d(n_angles, np.array([l_det, l_det]), np.ones(2, ), l_det, 1.0)
# #########################

fn = "phantom.png"
image = PIL.Image.open(fn).convert('L')

#convert to angle (180) map; modulo wraps 255 back to 0
map = (np.asarray(image) / 255 * n_angles).astype(int) % n_angles

# print unique domain angles
print('angles:', np.unique(map))
l_x = map.shape[0]  ##l_x times l_x is the size of the real space image

domain_maps = {angle: (map == angle)*2-1 for angle in np.unique(map) if angle == 139}
# n_sinograms = len(domain_maps)
n_sinograms = 1

if __name__ == "__main__":
    decision_queue = Queue_decision()

    # angles = np.linspace(0.0, np.pi, n_angles)

    #the following line will replaced by starting data, a sinogram with the right shape but only measured data non-zero
    # gt_A = np.array([projection_operator(x, phi, l_x) for x in range(l_det) for phi in range(n_angles)]).reshape(l_det * n_angles, -1)
    # domain_sinograms = {}
    # for domain_angle, domain_map in domain_maps.items():
    #     domain_sinograms[domain_angle] = (gt_A @ domain_map.ravel()).reshape(l_det, n_angles).T

    # name = r'C:\data\gitomo\run2-checkpoint1.yml'
    # data = Data(**load(open(name, 'r'), Loader=Loader))
    # positions = np.asarray(data.positions)
    # x, y = positions.T
    execution = SimpleEngine(measure_func=push_to_queue)

    # print(f'values: {min(data.scores), max(data.scores)}')

    # Define a gpCAM adaptive engine with initial parameters
    adaptive = ThresholdResolve(dimensionality=2,
                         grid_until_N=4,
                         parameter_bounds=[(0, l_x-1),  # NOTE: THIS -1 is important to avoid the projection weight not becoming empty
                                           (0, 180)],
                         hyperparameters=[1, 2],
                         hyperparameter_bounds=[[.1, 1e3],  # signal variance
                                                [.1, 1e4]],  # lengthscale
                         gp_opts=dict(gp_kernel_function=kernel,
                                      gp_mean_function=mean_func,
                                      #gp_noise_function=noise, # newstyle
                                      ),
                         # ask_opts=dict(vectorized=False),
                         acquisition_functions={'Binary Resolve': binary_resolve,
                                                'Projected Variance': variance,
                                                'Projected UCB': ucb})

    # adaptive = ProjectionOperatorGrid(parameter_bounds=[(0, l_x), (0, 180)])

    # x_data = np.empty((l_x ** 2, 2))
    # y_data = np.zeros((n_angles * l_det))

    # for i in range(l_x):
    #     for j in range(l_x):
    #         x_data[i+j*l_x] = np.array([float(i), float(j)])

    # v_data = 0.01 * np.ones((n_angles * l_det))

    # Construct a core server
    core = ZMQCore()
    core.set_adaptive_engine(adaptive)
    core.set_execution_engine(execution)
    # core.initialize_data(x_data, y_data, v_data)

    # Start the core server
    core.state = CoreState.Starting
    core.main()



import os
import threading
import time
from asyncio import events
from enum import Enum, auto
from pickle import UnpicklingError
from queue import Queue
from appdirs import user_state_dir

from loguru import logger
from yaml import dump

from .messages import FullDataRequest, FullDataResponse, PartialDataRequest, PartialDataResponse, StartRequest, \
    UnknownResponse, PauseRequest, StateRequest, GetParametersRequest, SetParameterRequest, GetParametersResponse, \
    SetParameterResponse, StopRequest, StateResponse, MeasureRequest, \
    MeasureResponse, ConnectRequest, ConnectResponse, ExceptionResponse, PushDataRequest, PushDataResponse, \
    GraphsResponse, ReplayResponse
from ..adaptive import Engine as AdaptiveEngine, Data
from ..execution import Engine as ExecutionEngine
from ..utils.logging import log_time

user_state_dir = user_state_dir('tsuchinoko','camera')

class CoreState(Enum):
    Connecting = auto()
    Inactive = auto()
    Starting = auto()
    Running = auto()
    Pausing = auto()
    Paused = auto()
    Resuming = auto()
    Stopping = auto()
    Restarting = auto()
    Exiting = auto()


SLEEP_FOR_FRESH_DATA_TIME = .1


class Core:
    def __init__(self,
                 execution_engine: ExecutionEngine = None,
                 adaptive_engine: AdaptiveEngine = None,
                 compute_metrics: bool = True):
        self.execution_engine = execution_engine
        self.adaptive_engine = adaptive_engine

        self.iteration = 0

        self._state = CoreState.Inactive
        self._exception_queue = Queue()
        self._forced_position_queue = Queue()
        self._forced_measurement_queue = Queue()
        self._has_fresh_data = True
        self.compute_metrics = compute_metrics
        self.checkpoint_template = 'checkpoint_{n}.yml'
        self.checkpoint_at = []
        self.pause_at = []
        self.stop_at = []
        self.exit_at = []
        self.compute_metrics_at = []

        self.data = Data()
        self._graphs = []

        self.experiment_thread = None

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        logger.info(f'Changing core state to {value}')
        self._state = value

    def set_execution_engine(self, engine: ExecutionEngine):
        self.execution_engine = engine

    def set_adaptive_engine(self, engine: AdaptiveEngine):
        self.adaptive_engine = engine

    def main(self, debug=False):
        loop = events.new_event_loop()  # <---- this ensures the current loop is replaced
        try:
            events.set_event_loop(loop)
            loop.set_debug(debug)
            return loop.run_until_complete(self._main())
        finally:
            try:
                # _cancel_all_tasks(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
            finally:
                events.set_event_loop(None)
                loop.close()

    async def _main(self, min_response_sleep=.1):
        while self.state != CoreState.Exiting:

            if self.state == CoreState.Running:
                pass
                # await sleep(min_response_sleep)  # short-circuit case
            elif self.state == CoreState.Starting:
                if not len(self.data):
                    self.data = Data(dimensionality=self.adaptive_engine.dimensionality)
                self.adaptive_engine.reset()
                self.experiment_thread = threading.Thread(target=self.experiment_loop, args=())  # must hold ref
                self.experiment_thread.start()
                self.state = CoreState.Running

            elif self.state == CoreState.Inactive:
                pass
                # await sleep(min_response_sleep)

            elif self.state == CoreState.Paused:
                pass
                # await sleep(min_response_sleep)

            elif self.state == CoreState.Pausing:
                self.state = CoreState.Paused

            elif self.state == CoreState.Resuming:
                self.state = CoreState.Running

            elif self.state == CoreState.Stopping:
                self.state = CoreState.Inactive
                self.data = Data()
                # await sleep(min_response_sleep)

            if self.state not in [CoreState.Stopping, CoreState.Exiting, CoreState.Resuming, CoreState.Restarting]:
                await self.notify_clients()

    def experiment_loop(self):
        while True:
            if self.state == CoreState.Running:
                logger.info(f'Iteration: {self.data._completed_iterations}, Data count: {len(self.data)}')
                if self.data._completed_iterations in self.checkpoint_at:
                    self.save_checkpoint()
                if self.data._completed_iterations in self.pause_at:
                    self.state = CoreState.Pausing
                    continue
                if self.data._completed_iterations in self.stop_at:
                    self.state = CoreState.Stopping
                    return
                if self.data._completed_iterations in self.exit_at:
                    self.state = CoreState.Exiting
                    return
                try:
                    self.experiment_iteration()
                except Exception as ex:
                    self._exception_queue.put(ex)
                    self.state = CoreState.Pausing
                    logger.exception(ex)
            elif self.state in [CoreState.Stopping, CoreState.Inactive, CoreState.Exiting]:
                return
            else:
                time.sleep(.1)

    def experiment_iteration(self):
        with self.data.iteration():
            if self._has_fresh_data:
                with log_time('getting position', cumulative_key='getting position'):
                    position = self.execution_engine.get_position()
                    logger.info(f'position: {position}')
                    if position is None:
                        position = [0] * self.data.dimensionality
                    position = tuple(position)
                if self._forced_position_queue.empty():
                    with log_time('getting targets', cumulative_key='getting targets'):
                        targets = self.adaptive_engine.request_targets(position)
                    logger.info(f'targets: {targets}')
                else:
                    targets = [self._forced_position_queue.get()]

            if self._forced_measurement_queue.empty():
                if self._has_fresh_data:
                    with log_time('updating targets', cumulative_key='updating targets'):
                        self.execution_engine.update_targets(targets)
                    self._has_fresh_data = False
                with log_time('getting measurements', cumulative_key='getting measurements'):
                    new_measurements = self.execution_engine.get_measurements()
                logger.info(f'new measurements: {new_measurements}')
            else:
                new_measurements = [self._forced_measurement_queue.get()]
            if len(new_measurements):
                self._has_fresh_data = True
                with log_time('stashing new measurements', cumulative_key='injecting new measurements'):
                    self.data.inject_new(new_measurements)
                with log_time('updating engine with new measurements', cumulative_key='updating engine with new measurements'):
                    self.adaptive_engine.update_measurements(self.data)
                if self.compute_metrics or len(self.data) in self.compute_metrics_at:
                    with log_time('updating metrics', cumulative_key='updating metrics'):
                        self.adaptive_engine.update_metrics(self.data)
            else:
                time.sleep(SLEEP_FOR_FRESH_DATA_TIME)
            if self._has_fresh_data:
                with log_time('training', cumulative_key='training'):
                    self.adaptive_engine.train()
            else:
                logger.info('Current data is stale. Waiting for an update with fresh data.')

    async def notify_clients(self):
        ...

    @property
    def graphs(self):
        execution_graphs = getattr(self.execution_engine, 'graphs', []) or []
        adaptive_graphs = getattr(self.adaptive_engine, 'graphs', []) or []
        return execution_graphs + adaptive_graphs + self._graphs

    @graphs.setter
    def graphs(self, graphs):
        raise NotImplementedError('Updating graphs on server not supported yet.')

    def update_graph(self, new_graph):
        execution_graphs = getattr(self.execution_engine, 'graphs', []) or []
        adaptive_graphs = getattr(self.adaptive_engine, 'graphs', []) or []
        self_graphs = self._graphs

        for graph_list in [execution_graphs, adaptive_graphs, self_graphs]:
            for i, old_graph in enumerate(graph_list):
                if old_graph.id == new_graph.id:
                    graph_list[i] = new_graph
                    return
        else:
            raise ValueError('Graph not found in graphs lists.')

    def initialize_data(self, x, y, v):
        with log_time('updating engine with initial measurements'):
            self.data = Data(dimensionality=len(x[0]), positions=x, scores=y, variances=v)
            self.adaptive_engine.update_measurements(self.data)

    def save_checkpoint(self, directory=user_state_dir):
        checkpoint_file_path = os.path.join(directory,
                                            self.checkpoint_template.format(n=self.data._completed_iterations))
        os.makedirs(os.path.dirname(checkpoint_file_path), exist_ok=True)
        dump(self.data.as_dict(), open(checkpoint_file_path, 'w'))


class ZMQCore(Core):
    def __init__(self, *args, **kwargs):
        super(ZMQCore, self).__init__(*args, **kwargs)
        # self.start_server()
        self.context = None
        self.poller = None

    def start_server(self):
        import zmq
        from zmq.asyncio import Context, Poller
        self.poller = Poller()
        self.context = Context()
        socket = self.context.socket(zmq.REP)
        socket.bind("tcp://*:5555")
        self.poller.register(socket, zmq.POLLIN)

    def respond_FullDataRequest(self, request):
        with self.data.r_lock():
            return FullDataResponse(self.data.as_dict())

    def respond_PartialDataRequest(self, request):
        if self.data and request.iteration <= len(self.data) and self.state == CoreState.Running:
            with self.data.r_lock():
                partial_data = self.data[request.iteration:]
            return PartialDataResponse(partial_data.as_dict(), request.iteration)
        else:
            return StateResponse(self.state, self.compute_metrics)

    def respond_PushDataRequest(self, request):
        self.data = Data(**request.data)
        return PushDataResponse()

    def respond_StartRequest(self, request):
        if self.state == CoreState.Paused:
            self.state = CoreState.Resuming
        elif self.state == CoreState.Inactive:
            self.state = CoreState.Starting
        return StateResponse(self.state, self.compute_metrics)

    def respond_StopRequest(self, request):
        self.state = CoreState.Stopping
        self.experiment_thread.join()
        return StateResponse(self.state, self.compute_metrics)

    def respond_ExitRequest(self, request):
        self.state = CoreState.Exiting
        return StateResponse(self.state, self.compute_metrics)

    def respond_PauseRequest(self, request):
        self.state = CoreState.Pausing
        return StateResponse(self.state, self.compute_metrics)

    def respond_StateRequest(self, request):
        if not self._exception_queue.empty():
            return ExceptionResponse(self._exception_queue.get())
        else:
            return StateResponse(self.state, self.compute_metrics)

    def respond_GetParametersRequest(self, request):
        return GetParametersResponse(self.adaptive_engine.parameters.saveState())

    def respond_SetParameterRequest(self, request):
        self.adaptive_engine.parameters.child(*request.child_path).setValue(request.value)
        return SetParameterResponse(True)

    def respond_MeasureRequest(self, request):
        self._forced_position_queue.put(request.position)
        return MeasureResponse(True)

    def respond_ConnectRequest(self, request):
        return ConnectResponse(self.state, self.compute_metrics)

    def respond_PullGraphsRequest(self, request):
        return GraphsResponse(self.graphs)

    def respond_PushGraphsRequest(self, request):
        for graph in request.graphs:
            try:
                self.update_graph(graph)
            except ValueError as ex:
                return ExceptionResponse("Graph ID not found in server's graphs.")
        # self.graphs = request.graphs
        return StateResponse(self.state, self.compute_metrics)

    def respond_SetComputeMetricsRequest(self, request):
        self.compute_metrics = request.compute_metrics
        return StateResponse(self.state, self.compute_metrics)

    def respond_ReplayRequest(self, request):
        self._forced_measurement_queue.queue.clear()
        self._forced_position_queue.queue.clear()

        for position in request.positions:
            self._forced_position_queue.put(position)
        for measurement in request.measurements:
            self._forced_measurement_queue.put(measurement)
        logger.critical(f'Queue lengths: {len(self._forced_measurement_queue.queue)} {len(self._forced_position_queue.queue)}')
        return ReplayResponse(True)

    async def notify_clients(self):
        import zmq
        if not self.poller:
            self.start_server()

        sockets = dict(await self.poller.poll(timeout=.1))
        for socket in sockets:
            try:
                request = await socket.recv_pyobj(zmq.NOBLOCK)
            except (zmq.ZMQError, zmq.error.Again) as ex:
                logger.exception(ex)
            except UnpicklingError as ex:
                logger.exception(ex)
                logger.critical('The above error prevented unpacking data from the client.')
            else:
                if not request:
                    time.sleep(.1)
                    continue

                logger.info(f"Received request: {request}")
                with log_time('preparing response', cumulative_key='preparing response'):
                    responder = getattr(self, f'respond_{request.__class__.__name__}', None)
                    if responder:
                        try:
                            response = responder(request)
                        except Exception as ex:
                            response = ExceptionResponse(ex)
                    else:
                        response = UnknownResponse()

                logger.info(f'Sending response: {response}')
                await socket.send_pyobj(response)

                if isinstance(response, UnknownResponse):
                    logger.exception(ValueError(f'Unknown request received: {request}'))
                    time.sleep(.1)

    def exit_later(self):
        self.state = CoreState.Exiting

    def exit(self):
        self.exit_later()
        self.experiment_thread.join()



import subprocess
import sys

import numpy as np
from scipy import ndimage

from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.core import ZMQCore
from tsuchinoko.execution.simple import SimpleEngine

install_warning = """
ATTENTION!
This demo depends on a package that is not included in the Tsuchinoko installation.
As a convenience, this demo will install the required dependencies for you.
"""

try:
    from perlin_noise import PerlinNoise
except ImportError:
    print(install_warning)
    response = input('Would you like to install perlin-noise [y/(n)]? ')
    if response != 'y':
        print('Aborting...')
        sys.exit(1)
    subprocess.check_call([sys.executable, "-m", "pip", "install", 'perlin-noise'])
    from perlin_noise import PerlinNoise



# More useful for display purposes; PerlinNoise interpolates internally and accepts float positions in range [0,1)
noise = PerlinNoise(octaves=8)
# size = 10
# shape = (size, size, size)
# indices = np.transpose(np.indices(shape)/size,(3,2,1,0)).reshape((np.prod(shape), 3))
# noise_map = map(noise, indices)




execution = SimpleEngine(measure_func=lambda position: (position, noise(position), 1, {}))


dimensionality = 5
# Define a gpCAM adaptive engine with initial parameters
adaptive = GPCAMInProcessEngine(dimensionality=dimensionality,
                                parameter_bounds=[(0, 10)] * dimensionality,
                                hyperparameters=[255] + [100] * dimensionality,
                                hyperparameter_bounds=[(0, 1e5)] * (dimensionality+1))

# Construct a core server
core = ZMQCore()
core.set_adaptive_engine(adaptive)
core.set_execution_engine(execution)

if __name__ == '__main__':
    # Start the core server
    core.main()



from functools import cached_property
from itertools import product
import time

import numpy as np
from pyqtgraph.parametertree.parameterTypes import SimpleParameter, ListParameter, GroupParameter

from tsuchinoko.adaptive import Engine, Data
from tsuchinoko.graphs.common import Variance, Score


def unit_cells(dims=2, slow_dim=-1, sort=True):
    visited = list(product(range(2), repeat=dims))
    # visited = [(0, 0), (1, 0), (0, 1), (1, 1)]
    yield from visited
    shift = .5
    diagonal = True
    while True:
        next_sites = []
        for site in visited:
            # print('site:', site)
            if diagonal and not np.any(np.asarray(site)==1):
                next_sites.append(tuple(np.asarray(site)+shift))
                # print('new1:', np.asarray(site)+shift)
            elif not diagonal:
                if site[slow_dim] == 0 and np.all(np.asarray(site)<1):  # right (fast)
                    for dim in range(dims):
                        if dim == slow_dim % dims: continue
                        shifts = np.zeros((dims,))
                        shifts[dim] = shift
                        next = tuple(np.asarray(site) + shifts)
                        next_sites.append(next)
                        # print('new2:', np.asarray(site) + shifts)
                if site[slow_dim] < 1:  # up (slow)
                    shifts = np.zeros((dims,))
                    shifts[slow_dim] = shift
                    next = tuple(np.asarray(site) + shifts)
                    next_sites.append(next)
                    # print('new3:', np.asarray(site) + shifts)

        yield from next_sites
        visited.extend(next_sites)
        if sort:
            visited = sorted(visited, key=lambda site: site[slow_dim]*10000+sum(site)-site[slow_dim])

        if not diagonal:
            shift /= 2
        diagonal = not diagonal


class Grid(Engine):
    dimensionality: int = None

    def __init__(self, parameter_bounds):
        self.dimensionality = len(parameter_bounds)

        for i in range(self.dimensionality):
            for j, edge in enumerate(['min', 'max']):
                self.parameters[('bounds', f'axis_{i}_{edge}')] = parameter_bounds[i][j]

        self.reset()
        self.graphs = [Variance(), Score(), ]


    @cached_property
    def parameters(self):
        bounds_parameters = [SimpleParameter(title=f'Axis #{i + 1} {edge}', name=f'axis_{i}_{edge}', type='float')
                             for i in range(self.dimensionality) for edge in ['min', 'max']]

        parameters = [GroupParameter(name='bounds', title='Axis Bounds', children=bounds_parameters),]
        return GroupParameter(name='top', children=parameters)

    def update_measurements(self, data: Data):
        ...

    def request_targets(self, position, **kwargs):
        bounds = [[self.parameters[('bounds', f'axis_{i}_{edge}')]
                   for edge in ['min', 'max']]
                  for i in range(self.dimensionality)]
        mins, maxs = zip(*bounds)
        target = np.asarray(next(self.unit_cells)) * (np.asarray(maxs)-np.asarray(mins))+np.asarray(mins)
        # time.sleep(.1)
        return [target]

    def reset(self):
        self.unit_cells = unit_cells()

    def train(self):
        ...

    def update_metrics(self, data: Data):
        ...



from queue import Queue
from typing import Callable, Tuple, List

from loguru import logger

from . import Engine


class SimpleEngine(Engine):
    def __init__(self, measure_func=Callable[[Tuple[float]], Tuple[float]], gives_variance=True, gives_position=True, gives_metrics=True):
        super(SimpleEngine, self).__init__()

        self.measure_func = measure_func
        self.position = None
        self.targets = Queue()
        self.new_measurements = []
        self._gives_variance = gives_variance
        self._gives_position = gives_position
        self._gives_metrics = gives_metrics

    def update_targets(self,  targets: List[Tuple]):
        with self.targets.mutex:
            self.targets.queue.clear()

        for target in targets:
            self.targets.put(target)

    def get_position(self) -> Tuple:
        return self.position

    def get_measurements(self) -> List[Tuple]:

        while not self.targets.empty():
            self.position = tuple(self.targets.get())
            measurement = self.measure_func(self.position)
            if not self._gives_position and not self._gives_position and not self._gives_metrics:
                measurement = [measurement]
            if not self._gives_position:
                measurement = [self.position, *measurement]
            if not self._gives_variance:
                measurement = [*measurement, .1]
            if not self._gives_metrics:
                measurement = [*measurement, {}]
            self.new_measurements.append(tuple(measurement))
            self.targets.task_done()

        measurements = self.new_measurements
        self.new_measurements = []
        return measurements


class SimplestEngine(SimpleEngine):
    def __init__(self, measure_func=Callable[[Tuple[float]], Tuple[float]]):
        super().__init__(measure_func, gives_position=False, gives_variance=False, gives_metrics=False)



from typing import Callable

import numpy as np
from pyqtgraph import ImageView, PlotWidget, RectROI, ImageItem, PlotItem
from qtpy.QtCore import QPointF, Signal, QObject, QEvent, Qt, QSignalBlocker
from qtpy.QtWidgets import QAction, QWidget, QHBoxLayout, QVBoxLayout, QPushButton, QSizePolicy
from pyqtgraph import functions as fn, debug, Point

from tsuchinoko.widgets.displays import Configuration


# TODO: map imageitems into target coordinate domain


class RequestRelay(QObject):
    sigRequestMeasure = Signal(tuple)


request_relay = RequestRelay()


class BetterLayout(ImageView):
    # Replaces awkward gridlayout with more structured v/hboxlayouts, and removes useless buttons
    def __init__(self, *args, **kwargs):
        super(BetterLayout, self).__init__(*args, **kwargs)

        # Shrink LUT
        self.getHistogramWidget().setMinimumWidth(10)

        self._reset_layout()
        self._set_layout()

    def _set_layout(self, layout=None):
        # Replace the layout
        QWidget().setLayout(self.ui.layoutWidget.layout())
        if layout is not None:
            self.ui.layoutWidget.setLayout(layout)
        else:
            self.ui.layoutWidget.setLayout(self.ui.outer_layout)

    def _reset_layout(self):
        self.ui.outer_layout = QHBoxLayout()
        self.ui.left_layout = QVBoxLayout()
        self.ui.right_layout = QVBoxLayout()
        self.ui.outer_layout.addLayout(self.ui.left_layout)
        self.ui.outer_layout.addLayout(self.ui.right_layout)
        for layout in [self.ui.outer_layout, self.ui.left_layout, self.ui.right_layout]:
            layout.setContentsMargins(0, 0, 0, 0)
            layout.setSpacing(0)

        self.ui.left_layout.addWidget(self.ui.graphicsView)
        self.ui.right_layout.addWidget(self.ui.histogram)

        # Must keep the roiBtn around; ImageView expects to be able to check its state
        self.ui.roiBtn.setParent(self)
        self.ui.roiBtn.hide()


class BetterButtons(BetterLayout):
    def __init__(self, *args, **kwargs):
        super(BetterButtons, self).__init__(*args, **kwargs)

        # Setup axes reset button
        self.resetAxesBtn = QPushButton("Reset Axes")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.resetAxesBtn.sizePolicy().hasHeightForWidth())
        self.resetAxesBtn.setSizePolicy(sizePolicy)
        self.resetAxesBtn.setObjectName("resetAxes")
        self.ui.right_layout.addWidget(self.resetAxesBtn)
        self.resetAxesBtn.clicked.connect(self.autoRange)

        # Setup LUT reset button
        self.resetLUTBtn = QPushButton("Reset LUT")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.resetLUTBtn.sizePolicy().hasHeightForWidth())
        # self.resetLUTBtn.setSizePolicy(sizePolicy)
        # self.resetLUTBtn.setObjectName("resetLUTBtn")
        self.ui.right_layout.addWidget(self.resetLUTBtn)
        self.resetLUTBtn.clicked.connect(self.autoLevels)


class AspectRatioLock(BetterLayout):
    def __init__(self, *args, lock_aspect=True, **kwargs):
        super().__init__(*args, **kwargs)
        self._aspect_ratio_button = QPushButton('Lock Aspect')
        self._aspect_ratio_button.setCheckable(True)
        self.ui.right_layout.addWidget(self._aspect_ratio_button)
        self._aspect_ratio_button.setChecked(lock_aspect)
        self._aspect_ratio_button.toggled.connect(self.toggle_lock_aspect)

    def toggle_lock_aspect(self):
        if self._aspect_ratio_button.isChecked():
            self.view.setAspectLocked(True)
        else:
            self.view.setAspectLocked(False)


class ComposableItemImageView(ImageView):
    """
    Used to compose together different image view mixins that may use different ItemImage subclasses.
    See LogScaleIntensity, LogScaleImageItem, ImageViewHistogramOverflowFIx, ImageItemHistorgramOverflowFix.
    Note that any imageItem named argument passed into the ImageView mixins above will discard the item and instead
    create a composition of imageItem_bases with their respective ImageItem class.
    """

    imageItem_bases = tuple()


class LogScaleImageItem(ImageItem):
    def __init__(self, *args, **kwargs):
        super(LogScaleImageItem, self).__init__(*args, **kwargs)
        self.logScale = True

    def render(self):
        # Convert data to QImage for display.

        profile = debug.Profiler()
        if self.image is None or self.image.size == 0:
            return
        if isinstance(self.lut, Callable):
            lut = self.lut(self.image)
        else:
            lut = self.lut

        if self.logScale:
            image = self.image + 1
            with np.errstate(invalid="ignore", divide='ignore'):
                image = image.astype(float)
                np.log(image, where=image >= 0, out=image)  # map to 0-255
        else:
            image = self.image

        if self.autoDownsample:
            # reduce dimensions of image based on screen resolution
            o = self.mapToDevice(QPointF(0, 0))
            x = self.mapToDevice(QPointF(1, 0))
            y = self.mapToDevice(QPointF(0, 1))
            w = Point(x - o).length()
            h = Point(y - o).length()
            if w == 0 or h == 0:
                self.qimage = None
                return
            xds = max(1, int(1.0 / w))
            yds = max(1, int(1.0 / h))
            axes = [1, 0] if self.axisOrder == "row-major" else [0, 1]
            image = fn.downsample(image, xds, axis=axes[0])
            image = fn.downsample(image, yds, axis=axes[1])
            self._lastDownsample = (xds, yds)
        else:
            pass

        # if the image data is a small int, then we can combine levels + lut
        # into a single lut for better performance
        levels = self.levels
        if levels is not None and levels.ndim == 1 and image.dtype in (np.ubyte, np.uint16):
            if self._effectiveLut is None:
                eflsize = 2 ** (image.itemsize * 8)
                ind = np.arange(eflsize)
                minlev, maxlev = levels
                levdiff = maxlev - minlev
                levdiff = 1 if levdiff == 0 else levdiff  # don't allow division by 0
                if lut is None:
                    efflut = fn.rescaleData(ind, scale=255.0 / levdiff, offset=minlev, dtype=np.ubyte)
                else:
                    lutdtype = np.min_scalar_type(lut.shape[0] - 1)
                    efflut = fn.rescaleData(
                        ind,
                        scale=(lut.shape[0] - 1) / levdiff,
                        offset=minlev,
                        dtype=lutdtype,
                        clip=(0, lut.shape[0] - 1)
                    )
                    efflut = lut[efflut]

                self._effectiveLut = efflut
            lut = self._effectiveLut
            levels = None

        # Assume images are in column-major order for backward compatibility
        # (most images are in row-major order)

        if self.axisOrder == "col-major":
            image = image.transpose((1, 0, 2)[: image.ndim])

        if self.logScale:
            with np.errstate(invalid="ignore"):
                levels = np.log(np.add(levels, 1))
            levels[0] = np.nanmax([levels[0], 0])

        argb, alpha = fn.makeARGB(image, lut=lut, levels=levels)
        self.qimage = fn.makeQImage(argb, alpha, transpose=False)


class LogScaleIntensity(BetterLayout, ComposableItemImageView):
    def __init__(self, *args, log_scale=True, **kwargs):
        # Composes a new type consisting of any ImageItem types in imageItem_bases with this classes's helper ImageItem
        # class (LogScaleImageItem)
        self.imageItem_bases += (LogScaleImageItem,)
        imageItem = type("DynamicImageItem", tuple(self.imageItem_bases), {})(np.zeros((1, 1)))
        if "imageItem" in kwargs:
            del kwargs["imageItem"]
        super(LogScaleIntensity, self).__init__(imageItem=imageItem, *args, **kwargs)

        self.logScale = log_scale

        # Setup log scale button
        self.logIntensityButton = QPushButton("Log Intensity")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.logIntensityButton.sizePolicy().hasHeightForWidth())
        self.logIntensityButton.setSizePolicy(sizePolicy)
        self.logIntensityButton.setObjectName("logIntensity")
        self.ui.right_layout.addWidget(self.logIntensityButton)
        self.logIntensityButton.setCheckable(True)
        self.setLogScale(self.logScale)
        self.logIntensityButton.clicked.connect(self._setLogScale)

    def _setLogScale(self, value):
        self.imageItem.logScale = value
        self.imageItem.qimage = None
        self.imageItem.update()
        self.getHistogramWidget().region.setBounds([0 if value else None, None])

    def setLogScale(self, value):
        self._setLogScale(value)
        self.logIntensityButton.setChecked(value)


class ViridisImageView(ImageView):
    def __init__(self, *args, **kwargs):
        super(ViridisImageView, self).__init__(*args, **kwargs)

        # Use Viridis by default
        self.setPredefinedGradient("viridis")


class BetterAutoLUTRangeImageView(ImageView):
    def __init__(self, *args, **kwargs):
        super(BetterAutoLUTRangeImageView, self).__init__(*args, **kwargs)

    def quickMinMax(self, data):
        """
        Estimate the min/max values of *data* by subsampling. MODIFIED TO USE THE 99TH PERCENTILE instead of max.
        """
        if data is None:
            return 0, 0
        ax = np.argmax(data.shape)
        sl = [slice(None)] * data.ndim
        sl[ax] = slice(None, None, max(1, int(data.size // 1e6)))
        for axis_sl in sl:
            data = data[axis_sl]
        return [(
            np.nanpercentile(np.where(data > np.nanmin(data), data, np.nanmax(data)), 2),
            np.nanpercentile(np.where(data < np.nanmax(data), data, np.nanmin(data)), 98),
        )]


class ClickRequesterBase:
    def __init__(self, *args, **kwargs):
        super(ClickRequesterBase, self).__init__(*args, **kwargs)

        self.measure_action = QAction('Queue Measurement at Point')
        self.measure_action.triggered.connect(self.emit_measure_request)
        self._scene().contextMenu.append(self.measure_action)
        self._last_mouse_event_pos = None
        self._install_filter()

    def _install_filter(self):
        ...

    def _scene(self):
        ...

    def eventFilter(self, obj, ev):
        if ev.type() == QEvent.Type.MouseButtonPress:
            if ev.button() == Qt.MouseButton.RightButton:
                self._last_mouse_event_pos = ev.pos()
        ev.ignore()

        return False


class YInvert(ImageView):
    def __init__(self, *args, invert_y=False, **kwargs):
        if 'view' in kwargs:
            raise ValueError(f'Setting view is incompatible with this widget ({type(self)}')
        graph = PlotItem()
        super().__init__(*args, view=graph, **kwargs)
        graph.vb.invertY(invert_y)


class ClickRequester(ClickRequesterBase, ImageView):
    def _scene(self):
        return self.scene

    def _install_filter(self):
        self.ui.graphicsView.installEventFilter(self)

    def emit_measure_request(self, *_):
        app_pos = self._last_mouse_event_pos
        # map to local pos
        local_pos = self.view.vb.mapSceneToView(app_pos)
        request_relay.sigRequestMeasure.emit(local_pos.toTuple())


class ClickRequesterPlot(ClickRequesterBase, PlotWidget):
    def _install_filter(self):
        self.installEventFilter(self)

    def _scene(self):
        return self.sceneObj

    def emit_measure_request(self, *_):
        app_pos = self._last_mouse_event_pos
        # map to local pos
        local_pos = self.plotItem.vb.mapSceneToView(app_pos)
        request_relay.sigRequestMeasure.emit(local_pos.toTuple())


class DomainROI(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        bounds = [tuple(Configuration().parameter.child('bounds')[f'axis_{i}_{limit}']
                        for limit in ['min', 'max'])
                  for i in range(2)]
        self._domain_roi = RectROI((bounds[0][0], bounds[1][0]),
                                   (bounds[0][1] - bounds[0][0], bounds[1][1] - bounds[1][0]))
        self._domain_roi.sigRegionChangeFinished.connect(self.update_bounds)
        self.getView().addItem(self._domain_roi)
        bounds_param = Configuration().parameter.child('bounds')
        for child in bounds_param.children():
            child.sigValueChanged.connect(self.update_roi)

    def update_bounds(self):
        bounds_param = Configuration().parameter.child('bounds')
        for child in bounds_param.children():
            child.sigValueChanged.disconnect(self.update_roi)
        bounds_param[f'axis_0_min'] = self._domain_roi.pos().x()
        bounds_param[f'axis_1_min'] = self._domain_roi.pos().y()
        bounds_param[f'axis_0_max'] = self._domain_roi.pos().x() + self._domain_roi.size().x()
        bounds_param[f'axis_1_max'] = self._domain_roi.pos().y() + self._domain_roi.size().y()
        for child in bounds_param.children():
            child.sigValueChanged.connect(self.update_roi)

    def update_roi(self):
        bounds_param = Configuration().parameter.child('bounds')
        blocker = QSignalBlocker(self._domain_roi)
        self._domain_roi.setPos(bounds_param[f'axis_0_min'], bounds_param[f'axis_1_min'], update=False)
        self._domain_roi.setSize(bounds_param[f'axis_0_max'] - bounds_param[f'axis_0_min'],
                                 bounds_param[f'axis_1_max'] - bounds_param[f'axis_1_min'])



from qtpy.QtWidgets import QSlider, QWidget, QHBoxLayout, QLabel
from qtpy.QtCore import Signal


class ValueDoubleSlider(QWidget):
    sigFloatValueChanged = Signal(float)
    '''Shows current value'''
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.slider = DoubleSlider(*args, **kwargs)
        self.label = QLabel()
        self.label.setFixedWidth(40)

        self.slider.sigFloatValueChanged.connect(self.sigFloatValueChanged)
        self.slider.sigFloatValueChanged.connect(self.update_label)

        ## Explicitly wrap methods from DoubleSlider
        for m in ['value', 'setMinimum', 'setMaximum', 'setInterval']:
            setattr(self, m, getattr(self.slider, m))

        self.setLayout(QHBoxLayout())
        self.layout().addWidget(self.label)
        self.layout().addWidget(self.slider)
        self.update_label(self.slider.value())

    def update_label(self, value):
        self.label.setText(str(value))


class DoubleSlider(QSlider):
    sigFloatValueChanged = Signal(float)

    def __init__(self, *args, **kargs):
        super(DoubleSlider, self).__init__( *args, **kargs)
        self._min = 0
        self._max = 99
        self.interval = 1
        self.valueChanged.connect(self._emit_value)

    def setValue(self, value):
        index = round((value - self._min) / self.interval)
        return super(DoubleSlider, self).setValue(index)

    def value(self):
        return self.index * self.interval + self._min

    @property
    def index(self):
        return super(DoubleSlider, self).value()

    def setIndex(self, index):
        return super(DoubleSlider, self).setValue(index)

    def setMinimum(self, value):
        self._min = value
        self._range_adjusted()

    def setMaximum(self, value):
        self._max = value
        self._range_adjusted()

    def setInterval(self, value):
        # To avoid division by zero
        if not value:
            raise ValueError('Interval of zero specified')
        self.interval = value
        self._range_adjusted()

    def _range_adjusted(self):
        number_of_steps = int((self._max - self._min) / self.interval)
        super(DoubleSlider, self).setMaximum(number_of_steps)

    def _emit_value(self):
        self.sigFloatValueChanged.emit(self.value())


from dataclasses import dataclass
from functools import partial, lru_cache
import sys
from loguru import logger
import fvgp

import numpy as np
from scipy import sparse, ndimage, fftpack, linalg
from fvgp.gp import GP
from gpcam.gp_optimizer import GPOptimizer
from tsuchinoko.adaptive import Data
from tsuchinoko.adaptive.gpCAM_in_process import GPCAMInProcessEngine
from tsuchinoko.adaptive.grid import Grid
from tsuchinoko.core import ZMQCore, CoreState
from tsuchinoko.execution.simple import SimpleEngine
from tsuchinoko.graphs import Location
from tsuchinoko.graphs.common import Table, Image, ImageViewBlendROI, image_grid, GPCamPosteriorMean, \
    GPCamHyperparameterPlot, Score, Variance, GPCamAcquisitionFunction
from tsuchinoko.utils import threads
from tsuchinoko.graphs.specialized import ReconstructionGraph, ProjectionOperatorGraph, \
    SinogramSpaceGPCamAcquisitionFunction, ProjectionMask
import PIL


# #################### FROM KANUPRIYA ########################
#
# class Geometry2d(object):
#     """
#     Detector and object setup for parallel beam geometry.
#     """
#
#     def __init__(self, n_proj, voxel_shape, voxel_pixsize, detector_shape, detector_pixsize,
#                  cor_shift=None, step_size=1.0):
#         """
#         :param n_proj: int, number of projections
#         :param voxel_shape: int, (2,)
#         :param voxel_pixsize: float (2,)
#         :param detector_shape: int (1,)
#         :param det_pixsize: float (1,)
#         """
#         self.n_proj = n_proj
#         self.vox_shape = voxel_shape
#         self.vox_pix = voxel_pixsize
#         self.vox_size = self.vox_shape * self.vox_pix
#         self.n_vox = np.prod(self.vox_shape)
#         self.det_shape = detector_shape
#         self.det_pix = detector_pixsize
#         self.det_size = self.det_shape * self.det_pix
#         self.n_det = np.prod(self.det_shape)
#         self.vox_ds = np.array([1, 1])
#         if cor_shift is None:
#             self.cor_shift = np.zeros((n_proj, 2))
#         else:
#             if len(cor_shift.shape) == 2:
#                 assert (cor_shift.shape[0] == n_proj)
#                 assert (cor_shift.shape[1] == 2)
#                 self.cor_shift = cor_shift
#             elif len(cor_shift.shape) == 1:
#                 assert (np.size(cor_shift) == 2)
#                 self.cor_shift = np.tile(cor_shift, n_proj).reshape(n_proj, 2)
#             else:
#                 print('shape or size of cor_shift not valid')
#
#         self.step_size = step_size
#         self._voxel_detector_grid()
#
#     def _voxel_detector_grid(self):
#
#         # first set up voxel centers and origin
#         nx, ny = self.vox_shape
#         sx, sy = self.vox_size
#         x = np.linspace(-sx / 2, sx / 2, nx, endpoint=False) + 0.5
#         y = np.linspace(-sy / 2, sy / 2, ny, endpoint=False) + 0.5
#         x, y = np.meshgrid(x, y, indexing='ij')
#         self.vox_centers = np.array([x.ravel(), y.ravel()])
#         self.vox_origin = np.array([x.min(), y.min()])
#
#         # now set up detector grid
#         nx = self.det_shape
#         sx = self.det_size
#         xd = np.linspace(-sx / 2, sx / 2, nx, endpoint=False) + 0.5
#         y_source = -sy
#         y_det = sy
#
#         # source and detector positions are needed for ray based method
#         self.source_centers = np.array([xd.ravel(), y_source * np.ones((self.n_det,))])
#         self.det_centers = np.array([xd.ravel(), y_det * np.ones((self.n_det,))])
#
#
#
# ############################################################



def trapez(y,y0,w):
    return np.clip(np.minimum(y+1+w/2-y0, -y+1+w/2+y0),0,1)


def weighted_line(r0, c0, r1, c1, w, rmin=0, rmax=np.inf):
    # The algorithm below works fine if c1 >= c0 and c1-c0 >= abs(r1-r0).
    # If either of these cases are violated, do some switches.
    if abs(c1-c0) < abs(r1-r0):
        # Switch x and y, and switch again when returning.
        xx, yy, val = weighted_line(c0, r0, c1, r1, w, rmin=rmin, rmax=rmax)
        return (yy, xx, val)

    # At this point we know that the distance in columns (x) is greater
    # than that in rows (y). Possibly one more switch if c0 > c1.
    if c0 > c1:
        return weighted_line(r1, c1, r0, c0, w, rmin=rmin, rmax=rmax)

    # The following is now always < 1 in abs
    slope = (r1-r0) / (c1-c0)

    # Adjust weight by the slope
    w *= np.sqrt(1+np.abs(slope)) / 2

    # We write y as a function of x, because the slope is always <= 1
    # (in absolute value)
    x = np.arange(c0, c1+1, dtype=float)
    y = x * slope + (c1*r0-c0*r1) / (c1-c0)

    # Now instead of 2 values for y, we have 2*np.ceil(w/2).
    # All values are 1 except the upmost and bottommost.
    thickness = np.ceil(w/2)
    yy = (np.floor(y).reshape(-1,1) + np.arange(-thickness-1,thickness+2).reshape(1,-1))
    xx = np.repeat(x, yy.shape[1])
    vals = trapez(yy, y.reshape(-1,1), w).flatten()

    yy = yy.flatten()

    # Exclude useless parts and those outside of the interval
    # to avoid parts outside of the picture
    mask = np.logical_and.reduce((yy >= rmin, yy < rmax, vals > 0))

    return (yy[mask].astype(int), xx[mask].astype(int), vals[mask])


def projection_operator(x, phi, map_size, center=None, width=1, length=None):
    if not length:
        length = map_size * 2

    if not center:
        center = (map_size/2, map_size/2)

    # get a sampling over a slice through x, phi
    v = np.array([-np.cos(np.deg2rad(phi)), np.sin(np.deg2rad(phi))]) * (-x + map_size/2)
    c = np.array(center)
    f = np.array([-np.sin(np.deg2rad(phi)), -np.cos(np.deg2rad(phi))]) * length / 2

    f0 = v + c + f
    f1 = v + c - f

    x0, y0 = f0
    x1, y1 = f1

    # draw a 1px line (with AA) as a mask from f0 to f1
    # NOTE: this is correct when the beam width is the same as the x/y resolution; for other cases,
    # you could use scipy.ndimage.zoom()
    rows, cols, vals = weighted_line(y0, x0, y1, x1, w=0.25)
    xy_mask = (cols >= 0) & (cols < map_size) & (rows >= 0) & (rows < map_size)
    masked_cols, masked_rows, masked_vals = cols[xy_mask], rows[xy_mask], vals[xy_mask]
    (map_mask := np.zeros((map_size, map_size)))[masked_rows, masked_cols] = masked_vals

    return map_mask


def kernel(x1,x2,hps,obj):
    # print('kernel:', hps, x1, x2)
    # with log_time('kernel', cumulative_key='kernel'):
    A = obj.A
    #this kernel takes elements from the real space the prior covariance in sinogram space
    d = obj._get_distance_matrix(x1, x2)
    k = hps[0] * np.exp(-d**2/hps[1])
    if len(x1) == len(x2) == len(obj.x_data):
        kernel = linalg.block_diag(*[A @ k @ A.T]*n_sinograms)
        # kernel = A @ k @ A.T
    elif len(x1) == len(x2) != len(obj.x_data):
        kernel = k
    elif len(x1) != len(x2):
        kernel = np.repeat(A @ k,n_sinograms,0)
        # kernel = A @ k
    return kernel


def noise(x,hps,obj):
    A = obj.A
    #this function takes elements from the real space and returns noise (co)variances in sinogram space
    n = A @ (np.zeros((len(x))) + 0.01)
    return np.diag(n)


def mean_func(x,hps,obj):
    A = obj.A
    #this function takes elements from the real space and returns prior mean values in sinogram space
    # m = A @ np.zeros(len(x),)
    if len(x) == len(obj.x_data):
        m = np.repeat(A @ np.zeros(len(x)), n_sinograms)
    else:
        m = np.zeros((len(x),))
    return m


def cost(origin, x, arguments=None):
    print('cost:', origin, x)
    cost_x = 1 / 8  # cost is 1/velocity
    cost_phi = 1 / 30
    exposure_time = 1
    estimated_total_experiment_time = 60*60*1  # in seconds
    x_origin, phi_origin = origin
    cost = [(exposure_time + max(abs(ix - x_origin) * cost_x, abs(iphi - phi_origin) * cost_phi))/estimated_total_experiment_time for
            ix, iphi in x]
    return cost


def bilinear_sample(pos, data):
    # pos = np.random.random((2,)) * np.array([32, 180])
    # print(f'measuring: x={pos[1]:.1f} y={pos[0]:.1f}')
    return pos, [ndimage.map_coordinates(sino, [[pos[1]], [pos[0]]], order=1)[0] for sino in data.values()], [.001] * n_sinograms, {}


class BackgroundTraining(GPCAMInProcessEngine):
    def __init__(self, start_training_at:int = 40, *args, **kwargs):
        self.training_thread = None
        self.start_training_at = start_training_at
        super().__init__(*args, **kwargs)

    def train(self):
        if not self.training_thread or self.training_thread.done:
            if len(self.optimizer.y_data) >= self.start_training_at:
                # pull values from optimizer
                x, y, v, A = self.optimizer.x_data.copy(), self.optimizer.y_data.copy(), self.optimizer.variances.copy(), self.optimizer.A.copy()

                # pull parameters
                hyperparameters_bounds = np.asarray([[self.parameters[('hyperparameters', f'hyperparameter_{i}_{edge}')]
                             for edge in ['min', 'max']]
                            for i in range(self.num_hyperparameters)])
                hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                            for i in range(self.num_hyperparameters)])
                parameter_bounds = np.asarray([[self.parameters[('bounds', f'axis_{i}_{edge}')]
                                                for edge in ['min', 'max']]
                                               for i in range(self.dimensionality)])

                self.training_thread = threads.QThreadFuture(self._background_train,
                                                             x,
                                                             y,
                                                             v,
                                                             A,
                                                             parameter_bounds,
                                                             hyperparameters,
                                                             hyperparameters_bounds,
                                                             {'method': 'global'})
                self.training_thread.start()

        return True

    def _background_train(self, x, y, v, A, parameter_bounds, hyperparameters, hyperparameter_bounds, training_kwargs):
        logger.info('Training asynchronously...')

        optimizer = GPOptimizer(self.dimensionality, parameter_bounds)
        optimizer.A = A
        optimizer.tell(x, y, v)
        optimizer.init_gp(hyperparameters, **self.gp_opts.copy())
        optimizer.init_cost(cost)
        optimizer.train(hyperparameter_bounds, hyperparameters, **training_kwargs)

        self.optimizer.hyperparameters = optimizer.hyperparameters
        logger.info(f'Hyperparameters set from asynchronous training: {optimizer.hyperparameters}')


class ProjectionOperatorBuilderGPCamEngine(BackgroundTraining):
    default_retrain_locally_at = tuple()
    default_retrain_globally_at = tuple() #82

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.graphs = [
            # GPCamPosteriorCovariance(),
            #            SinogramSpaceGPCamAcquisitionFunction(shape=(32, 32), real_space_bounds=((0, l_x), (0, l_x))),
                       ReconstructionGraph(shape=(l_x, l_x)),
                       ProjectionMask(shape=(l_x, l_x)),
                       ProjectionOperatorGraph(),
                       # GPCamAverageCovariance(),
                       GPCamHyperparameterPlot(),
                       Table(),
                       # Variance(),
                       Score()
        ]

    def init_optimizer(self):
        super().init_optimizer()
        self.optimizer.init_cost(cost)

    def update_measurements(self, data: Data):
        with data.r_lock():  # quickly grab values within lock before passing to optimizer
            positions = data.positions.copy()
            scores = data.scores.copy()
            variances = data.variances.copy()

        real_space_positions = np.mgrid[:l_x,:l_x].T.reshape(-1, 2)
        scores = np.array(scores).ravel()
        variances = np.array(variances).ravel()

        length_diff = len(positions) - self._positions_seen
        if length_diff:
            for position in positions[-length_diff:]:
                A_vec = projection_operator(*position, l_x).reshape(1, l_x **2)
                self.optimizer.A = np.vstack([self.optimizer.A, A_vec])
                self._positions_seen += 1
                # self.optimizer.positions = np.vstack([self.optimizer.positions, position])

        self.optimizer.tell(np.asarray(real_space_positions), np.asarray(scores), np.asarray(variances))
        if not self.optimizer.gp_initialized:
            hyperparameters = np.asarray([self.parameters[('hyperparameters', f'hyperparameter_{i}')]
                                          for i in range(self.num_hyperparameters)])
            opts = self.gp_opts.copy()
            # TODO: only fallback to numpy when packaged as an app
            if sys.platform == 'darwin':
                opts['compute_device'] = 'numpy'

            self.init_gp(hyperparameters, **opts)

    def reset(self):
        super().reset()
        self.optimizer.A = np.zeros((0, l_x ** 2))
        self.optimizer.positions = np.zeros((0, 2))
        self._positions_seen = 0
        if self.graphs:
            self.graphs[0].last_recon = None


class EmptyOptimizer():
    pass


class ProjectionOperatorGrid(Grid):
    def __init__(self, *args, **kwargs):
        self.optimizer = EmptyOptimizer()
        super().__init__(*args, **kwargs)
        self.graphs = [ReconstructionGraph(),
                       Table(),
                       Score()
                       ]

    def update_measurements(self, data: Data):
        with data.r_lock():  # quickly grab values within lock before passing to optimizer
            positions = data.positions.copy()

        length_diff = len(positions) - self._positions_seen
        if length_diff:
            for position in positions[-length_diff:]:
                A_vec = projection_operator(*position, l_x).reshape(1, l_x **2)
                self.optimizer.A = np.vstack([self.optimizer.A, A_vec])
                self._positions_seen += 1

    def update_metrics(self, data: Data):
        for graph in self.graphs:
            try:
                graph.compute(data, self)
            except Exception as ex:
                logger.exception(ex)

    def reset(self):
        super().reset()
        self.optimizer.A = np.zeros((0, l_x ** 2))
        self._positions_seen = 0
        if self.graphs:
            self.graphs[0].last_recon = None


class GridFirst(ProjectionOperatorBuilderGPCamEngine):
    def __init__(self, grid_until_N:int, parameter_bounds, *args, **kwargs):
        self.grid_until_N = grid_until_N
        self.grid_engine = Grid(parameter_bounds=parameter_bounds)
        super().__init__(*args, parameter_bounds=parameter_bounds, **kwargs)

    def request_targets(self, position):
        if len(self.optimizer.y_data) < self.grid_until_N:
            self.last_position = position
            return self.grid_engine.request_targets(position)
        else:
            return super().request_targets(position)

    def reset(self):
        super().reset()

        self.grid_engine.reset()



if __name__ == "__main__":
    n_angles = 180  # number of angles
    l_det = 32  # sampling for ground truth sinograms

    # #### FOR KANUPRIYA's ####
    # geom2d = Geometry2d(n_angles, np.array([l_det, l_det]), np.ones(2, ), l_det, 1.0)
    # #########################

    fn = "phantom.png"
    image = PIL.Image.open(fn).convert('L')

    # convert to angle (180) map; modulo wraps 255 back to 0
    map = (np.asarray(image) / 255 * n_angles).astype(int) % n_angles

    # print unique domain angles
    print('angles:', np.unique(map))
    l_x = map.shape[0]  ##l_x times l_x is the size of the real space image

    domain_maps = {angle: map == angle for angle in np.unique(map) if angle == 139}
    n_sinograms = len(domain_maps)

    angles = np.linspace(0.0, np.pi, n_angles)

    #the following line will replaced by starting data, a sinogram with the right shape but only measured data non-zero
    gt_A = np.array([projection_operator(x, phi, l_x) for x in range(l_det) for phi in range(n_angles)]).reshape(l_det * n_angles, -1)
    domain_sinograms = {}
    for domain_angle, domain_map in domain_maps.items():
        domain_sinograms[domain_angle] = (gt_A @ domain_map.ravel()).reshape(l_det, n_angles).T

    execution = SimpleEngine(measure_func=partial(bilinear_sample, data=domain_sinograms))

    # Define a gpCAM adaptive engine with initial parameters
    adaptive = GridFirst(dimensionality=2,
                         grid_until_N=81,
                         parameter_bounds=[(0, l_x),
                                           (0, 180)],
                         hyperparameters=[9, .16],
                         hyperparameter_bounds=[[.1, 1e3],  # signal variance
                                                [.1, 1e4]],  # lengthscale
                         gp_opts=dict(gp_kernel_function=kernel,
                                      gp_mean_function=mean_func,
                                      # gp_noise_function=noise,
                                      ),
                         ask_opts=dict(vectorized=False))

    # adaptive = ProjectionOperatorGrid(parameter_bounds=[(0, l_x), (0, 180)])

    # x_data = np.empty((l_x ** 2, 2))
    # y_data = np.zeros((n_angles * l_det))

    # for i in range(l_x):
    #     for j in range(l_x):
    #         x_data[i+j*l_x] = np.array([float(i), float(j)])

    # v_data = 0.01 * np.ones((n_angles * l_det))

    # Construct a core server
    core = ZMQCore()
    core.set_adaptive_engine(adaptive)
    core.set_execution_engine(execution)
    # core.initialize_data(x_data, y_data, v_data)

    # Start the core server
    # core.state = CoreState.Starting
    core.main()



import numpy as np

from tsuchinoko.examples.gi_tomo_phantom import projection_operator, l_x, n_sinograms, domain_maps, bilinear_sample
from tsuchinoko.utils.zmq_queue import Queue_measure

measure_queue = Queue_measure()



if __name__ == "__main__":
    while True: # The loop that waits for new instructions...

        pos = measure_queue.get()  # Get measurement command
        print('received:', pos)
        pos[0]['value'] = bilinear_sample(pos[0]['position'], domain_maps)
        print('sending:', pos)
        measure_queue.publish(pos)  # Send new results for analysis




# this is just to mock an analysis element. It effectively just relay's the
# measured value here
from tsuchinoko.utils.zmq_queue import Queue_analyze

q = Queue_analyze()

if __name__ == '__main__':
    while True:  # The loop that waits for new instructions...

        data = q.get()  # Get analysis command

        print("'Analyzing' data:", data)

        q.publish(data)  # Send new analysis results to gpCAM



import importlib
import subprocess
import sys
from typing import List


def oxfordcomma(listed):
    if len(listed) == 0:
        return ''
    if len(listed) == 1:
        return listed[0]
    if len(listed) == 2:
        return listed[0] + ' and ' + listed[1]
    return ', '.join(listed[:-1]) + ', and ' + listed[-1]


def ensure_dependencies(deps: List[str]):
    install_warning = """
    ATTENTION!
    This demo depends on a package that is not included in the Tsuchinoko installation.
    As a convenience, Tsuchinoko may install the required dependencies for you.
    """

    try:
        from perlin_noise import PerlinNoise
    except ImportError:
        print(install_warning)
        response = input(f'Would you like to install {oxfordcomma(deps)} [y/(n)]? ')
        if response != 'y':
            print('Aborting...')
            sys.exit(1)
        subprocess.check_call([sys.executable, "-m", "pip", "install", *deps])


def check_dependencies(deps: List[str]) -> bool:
    for dep in deps:
        if dep in sys.modules:
            print(f"{dep!r} already in sys.modules")
        elif (spec := importlib.util.find_spec(dep)) is not None:
            pass
            # module = importlib.util.module_from_spec(spec)
            # sys.modules[dep] = module
            # spec.loader.exec_module(module)
            # print(f"{dep!r} has been imported")
        else:
            return False
    return True


#!/usr/bin/python3

import os
import time, datetime
import zmq
import numpy as np


class CustomQueue():

    def __init__(self, from_port, to_port, from_ip='localhost', to_ip='*', name='', save_dir='./', verbosity=4,
                 **kwargs):

        # Save these in case we want to check them later
        self.from_ip = from_ip
        self.from_port = from_port
        self.to_ip = to_ip
        self.to_port = to_port

        self.name = name
        self.save_dir = save_dir
        self.verbosity = verbosity
        self.kwargs = kwargs

        self.context = zmq.Context()

        self.msg('Starting queues...', 2)

        # To/Tell/Publish/Push/Producer
        # Act as a "server", sending data when requested
        self.msg('Starting push queue (server)', 3, 1)
        self.msg('IP: {}, port: {}'.format(to_ip, to_port), 3, 2)
        success, attempt, start_time = False, 0, time.time()
        while not success:
            attempt += 1
            try:
                self.to_socket = self.context.socket(zmq.PUSH)
                self.to_socket.bind("tcp://{}:{:d}".format(to_ip, to_port))
                success = True
                self.msg('Success (attempt {:d}, {:.1f} s)'.format(attempt, time.time() - start_time), 3, 2)
            except zmq.error.ZMQError:
                self.msg('Retrying (attempt {:d}, {:.1f} s)'.format(attempt, time.time() - start_time), 3, 2)
                time.sleep(1)

        # From/Ask/Subscribe/Pull/Consumer
        # Act as a "client", asking for data
        self.msg('Starting pull queue (client)', 3, 1)
        self.msg('IP: {}, port: {}'.format(from_ip, from_port), 3, 2)
        success, attempt, start_time = False, 0, time.time()
        while not success:
            attempt += 1
            try:
                self.from_socket = self.context.socket(zmq.PULL)
                self.from_socket.connect("tcp://{}:{:d}".format(from_ip, from_port))
                success = True
                self.msg('Connected to pull server (attempt {:d}, {:.1f} s)'.format(attempt, time.time() - start_time),
                         3, 2)
            except zmq.error.ZMQError:
                self.msg('Waiting for pull server (attempt {:d}, {:.1f} s)'.format(attempt, time.time() - start_time),
                         3, 2)
                time.sleep(1)

        self.msg('Started.', 3)

    def msg(self, txt, threshold=3, indent=0, indent_txt='  '):
        if self.verbosity >= threshold:
            indent = np.clip(indent, 0, 10)
            indent = indent_txt * indent
            print('ZMQ: {}> {}{}'.format(self.name, indent, txt))

    def now(self, str_format='%Y-%m-%d %H:%M:%S'):
        return time.strftime(str_format)

    def time_str(self, timestamp, str_format='%Y-%m-%d %H:%M:%S'):
        # time_tuple = time.gmtime(timestamp)
        # s = time.strftime(str_format, time_tuple)
        s = datetime.datetime.fromtimestamp(timestamp).strftime(str_format)
        return s

    def time_delta(self, first, second):
        diff = abs(second - first)
        if diff > 60 * 60:
            diff = '{:.1f} hours'.format(diff / (60 * 60))
        elif diff > 60:
            diff = '{:.1f} minutes'.format(diff / 60)
        elif diff > 0.1:
            diff = '{:.1f} s'.format(diff)
        else:
            diff = '{:.4f} s'.format(diff)

        if second < first:
            return '{} earlier'.format(diff)
        else:
            return '{} later'.format(diff)

    def reply(self):
        # Wait for next request from client
        message = self.to_socket.recv()
        self.msg('Received request: {}'.format(message))

        # Do some 'work'
        time.sleep(1)

        # Send reply back
        self.to_socket.send("REPLY")

    def ask(self):
        self.msg('Sending request', 3, 1)
        self.from_socket.send(b"ASK")
        message = self.from_socket.recv()
        self.msg('Received reply: {}'.format(message))

    def get(self, save=True, check_interrupted=False, force_load=False, flags:int=0):
        '''Get the current item being published.'''
        # message = self.from_socket.recv()

        self.msg('Getting data/command ({})'.format(self.now()), 4, 1)

        if force_load or (check_interrupted and self.interrupted()):
            self.msg('Loading data/command from disk...'.format(self.now()), 4, 1)
            data = self.load()

            if isinstance(data, (list, tuple, np.ndarray)):
                self.msg('Loaded: list length {}'.format(len(data)), 4, 2)
            else:
                self.msg('Loaded.', 4, 2)

        else:
            self.msg('Waiting for data/command ({})...'.format(self.now()), 4, 1)
            data = self.from_socket.recv_pyobj(flags=flags)

            if isinstance(data, (list, tuple, np.ndarray)):
                self.msg('Received: list length {}'.format(len(data)), 4, 2)
            else:
                self.msg('Received.', 4, 2)

            if save:
                np.save('{}/{}-received.npy'.format(self.save_dir, self.name), data, allow_pickle=True)

        return data

    def publish(self, data, save=True):
        # message = '{} {}'.format(self.name, data)
        # self.to_socket.send(message)

        self.msg('Sending data/command ({})...'.format(self.now()), 4, 1)
        self.to_socket.send_pyobj(data)
        self.msg('Sent.', 4, 2)

        if save:
            np.save('{}/{}-sent.npy'.format(self.save_dir, self.name), data, allow_pickle=True)

    def interrupted(self):

        received = '{}/{}-received.npy'.format(self.save_dir, self.name)

        if not os.path.exists(received):
            self.msg('No saved received data from prior work-cycle.', 4, 2)
            return False

        received = os.path.getmtime(received)
        sent = '{}/{}-sent.npy'.format(self.save_dir, self.name)
        if os.path.isfile(sent):
            sent = os.path.getmtime(sent)
        else:
            self.msg('No saved sent data from prior work-cycle.', 4, 2)
            return True

        if sent > received:
            # As expected, the file we sent is newer
            self.msg('Last work-cycle completed normally (based on file dates).', 4, 2)
            self.msg('Received: {}'.format(self.time_str(received)), 5, 3)
            self.msg('Sent: {} ({})'.format(self.time_str(sent), self.time_delta(received, sent)), 5, 3)
            return False
        else:
            self.msg('Last work-cycle INTERRUPTED (based on file dates).', 3, 2)
            self.msg('Received: {}'.format(self.time_str(received)), 3, 3)
            self.msg('Sent: {} ({})'.format(self.time_str(sent), self.time_delta(received, sent)), 3, 3)
            return True

    def load(self, stype='received'):
        data = np.load('{}/{}-{}.npy'.format(self.save_dir, self.name, stype), allow_pickle=True)
        return data

    def print_connections(self):
        txt = 'Push queue: IP: {}, port: {}\nPull queue: IP: {}, port: {}'.format(self.to_ip, self.to_port,
                                                                                  self.from_ip, self.from_port)
        self.msg(txt, 1, 0)

    def clear(self):
        '''Remove the saved filed.'''
        self.msg('Clearing queue saved files.', 2, 1)

        received = '{}/{}-received.npy'.format(self.save_dir, self.name)
        if os.path.exists(received):
            self.msg('Removing {}'.format(received), 3, 2)
            os.remove(received)
        else:
            self.msg('Received data oes not exist ({})'.format(received), 3, 2)

        sent = '{}/{}-sent.npy'.format(self.save_dir, self.name)
        if os.path.exists(sent):
            self.msg('Removing {}'.format(sent), 3, 2)
            os.remove(sent)
        else:
            self.msg('Sent data oes not exist ({})'.format(sent), 3, 2)


VERBOSITY = 4


# Simple example of three systems working in a ring/loop:
########################################
class OneQueue(CustomQueue):
    def __init__(self, from_port=5553, to_port=5551, from_ip='localhost', to_ip='*', name='One', verbosity=VERBOSITY,
                 **kwargs):
        super().__init__(from_port=from_port, to_port=to_port, from_ip=from_ip, to_ip=to_ip, name=name,
                         verbosity=verbosity, **kwargs)


class TwoQueue(CustomQueue):
    def __init__(self, from_port=5551, to_port=5552, from_ip='localhost', to_ip='*', name='Two', verbosity=VERBOSITY,
                 **kwargs):
        super().__init__(from_port=from_port, to_port=to_port, from_ip=from_ip, to_ip=to_ip, name=name,
                         verbosity=verbosity, **kwargs)


class ThreeQueue(CustomQueue):
    def __init__(self, from_port=5552, to_port=5553, from_ip='localhost', to_ip='*', name='Three', verbosity=VERBOSITY,
                 **kwargs):
        super().__init__(from_port=from_port, to_port=to_port, from_ip=from_ip, to_ip=to_ip, name=name,
                         verbosity=verbosity, **kwargs)


# Example of arbitrary-sized ring:
########################################
if False:
    # Servers
    c = [
        {'name': 'decision', 'ip': '127.0.0.1', 'port': 5551},
        {'name': 'measure', 'ip': '127.0.0.1', 'port': 5552},
        {'name': 'process', 'ip': '127.0.0.1', 'port': 5553},
        {'name': 'analyze', 'ip': '127.0.0.1', 'port': 5554},
    ]
    queues = []
    for i, server in enumerate(c):
        # Connect each process 'from' to the prior entry in the list, and their 'to' to the current entry.
        from_i = i - 1 if i > 0 else len(c) - 1
        to_i = i
        arguments = {'name': c[i]['name'], 'from_ip': c[from_i]['ip'], 'from_port': c[from_i]['port'],
                     'to_ip': c[to_i]['ip'], 'to_port': c[to_i]['port']}
        queues.append(arguments)

    # Then each process can be started using:
    # from CustomQueue import *
    # q = CustomQueue(**queues[0]) # Each process uses a different index in the queues list
    # while True:
    # data = q.get()
    # q.publish(data)

# Loop for Autonomous Experimentation (AE):
########################################
local = '127.0.0.1'

## pre-reIP (November 2021)
# xf11bm_ws1 = '10.11.128.201'
# xf11bm_ws2 = '10.11.128.202'
# xf11bm_ws3 = '10.11.128.203'
# xf11bm_ws4 = '10.11.128.204'
# xf11bm_ws5 = '10.11.128.205'
# xf11bm_ws6 = '10.11.128.206'
#
# xf11bm_srv1 = '10.11.128.1'
# xf11bm_gpu1 = '10.11.128.3'
# xf11bm_gpu2 = '10.11.128.5'

# xf12id_ws1 = '10.12.0.201'
# xf12id_ws2 = '10.12.0.202'
# xf12id_ws3 = '10.12.0.203'
# xf12id_srv1 = '10.12.0.4'
# xf12id_srv2 = '10.12.0.5'

## post-reIP (November 2021)
xf11bm_ws1 = '10.68.80.221'
xf11bm_ws2 = '10.68.80.222'
xf11bm_ws3 = '10.68.80.223'
xf11bm_ws4 = '10.68.80.224'
xf11bm_ws5 = '10.68.80.225'
xf11bm_ws6 = '10.68.80.226'

xf11bm_srv1 = '10.68.80.25'
xf11bm_gpu1 = '10.68.80.27'
xf11bm_gpu2 = '10.68.80.29'

# Server connections
# c = {
#     'decision': {'ip': xf11bm_srv1, 'port': 5551},
#     'measure': {'ip': xf11bm_ws1, 'port': 5552},
#     'analyze': {'ip': xf11bm_srv1, 'port': 5553},
# }

c = {
    'decision': {'ip': local, 'port': 5557},
    'measure': {'ip': local, 'port': 5558},
    'analyze': {'ip': local, 'port': 5559},
}

# c = {
#    'decision' : {'ip': xf12id_srv1, 'port': 5551} ,
#    'measure' : {'ip': xf12id_ws1, 'port': 5552} ,
#    'analyze' : {'ip': xf12id_srv1, 'port': 5553} ,
#    }


# c = {  'decision': {'ip': local, 'port': 5551},  'measure': {'ip': local, 'port': 5552},  'analyze': {'ip': local, 'port': 5553}  } # When all processes are on the same machine


class Queue_decision(CustomQueue):  # gpCAM
    def __init__(self, from_port=c['analyze']['port'], to_port=c['decision']['port'], from_ip=c['analyze']['ip'],
                 to_ip=c['decision']['ip'], name='decision', save_dir='./', verbosity=VERBOSITY, **kwargs):
        super().__init__(from_port=from_port, to_port=to_port, from_ip=from_ip, to_ip=to_ip, name=name,
                         save_dir=save_dir, verbosity=verbosity, **kwargs)

    def get(self, save=True, check_interrupted=False, force_load=False):
        return super().get(save=save, check_interrupted=check_interrupted, force_load=force_load)


class Queue_measure(CustomQueue):  # beamline
    def __init__(self, from_port=c['decision']['port'], to_port=c['measure']['port'], from_ip=c['decision']['ip'],
                 to_ip=c['measure']['ip'], name='measure', save_dir='./', verbosity=VERBOSITY, **kwargs):
        super().__init__(from_port=from_port, to_port=to_port, from_ip=from_ip, to_ip=to_ip, name=name,
                         save_dir=save_dir, verbosity=verbosity, **kwargs)

    def get(self, save=True, check_interrupted=False, force_load=False):
        return super().get(save=save, check_interrupted=check_interrupted, force_load=force_load)


class Queue_analyze(CustomQueue):  # SciAnalysis
    def __init__(self, from_port=c['measure']['port'], to_port=c['analyze']['port'], from_ip=c['measure']['ip'],
                 to_ip=c['analyze']['ip'], name='analyze', save_dir='./', verbosity=VERBOSITY, **kwargs):
        super().__init__(from_port=from_port, to_port=to_port, from_ip=from_ip, to_ip=to_ip, name=name,
                         save_dir=save_dir, verbosity=verbosity, **kwargs)

    def get(self, save=True, check_interrupted=False, force_load=False):
        return super().get(save=save, check_interrupted=check_interrupted, force_load=force_load)


class Queue_analyzeFix(Queue_analyze):  # SciAnalysis

    def __init__(self, from_port=c['measure']['port'], to_port=c['analyze']['port'], from_ip=c['measure']['ip'],
                 to_ip=c['analyze']['ip'], name='analyze', save_dir='./', verbosity=VERBOSITY, **kwargs):
        super().__init__(from_port=from_port, to_port=to_port, from_ip=from_ip, to_ip=to_ip, name=name,
                         save_dir=save_dir, verbosity=verbosity, **kwargs)

        self.list_lengths = []

    def get(self, save=True, check_interrupted=True, force_load=False):

        data = super().get(save=save, check_interrupted=check_interrupted, force_load=force_load)

        self.list_lengths.append(len(data))
        if len(self.list_lengths) >= 2:
            len_cur = self.list_lengths[-1]
            len_prev = self.list_lengths[-2]

            if len_cur == len_prev:
                self.msg("WARNING: Possible double-loop detected.", 3, 1)
                self.msg("Ignoring this command and waiting for next command instead.", 3, 2)
                self.get(save=save, check_interrupted=False, force_load=False)

        return data

########################################
# The usage would be:
########################################

# Inside gpCAM:
########################################
# from CustomQueue import *
# q = Queue_decision()

# if first_iteration:
## Send an initial signal to the beamline
# data = np.asarray([0])
# q.publish(data)

# while True: # The loop that waits for new instructions...

# data = q.get() # Get analysis results

## Do whatever work needs to be done
## time.sleep(1)

# q.publish(data) # Send new command to beamline


# Inside beamline user.py:
########################################
# try:
# measure_queue
# except NameError:
# from CustomQueue import *
# measure_queue = Queue_measure()

# while True: # The loop that waits for new instructions...

# data = measure_queue.get() # Get measurement command

## Do whatever work needs to be done
## time.sleep(1)

# measure_queue.publish(data) # Send new results for analysis


# Inside SciAnalysis autonomous.py:
########################################
# from CustomQueue import *
# q = Queue_analyze()

# while True: # The loop that waits for new instructions...

# data = q.get() # Get analysis command

## Do whatever work needs to be done
## time.sleep(1)

# q.publish(data) # Send new analysis results to gpCAM




